// Code by Anna Choromanska and Tony Jebara
// Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "quickmatrix.hpp"
#include "mex.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{   
    //Inputs
    unsigned int t, DD, d_small, n, k;
    double *XX; 
    unsigned int *YY; 
    double *theta;
    double *V;
    double *S;
    double *D_vec;
    double *V_in;
    double *S_in;
    double *D_vec_in;
    double LAMBDA;
    double xi;
    
	//Outputs
    double *ui;
    double *pzti;
    double *THETA_OUT;
       
    //Inputs
    //Allocate memory and assign input pointer
    t = (unsigned int)(mxGetScalar(prhs[0]));
    DD = (unsigned int)(mxGetScalar(prhs[1]));
    d_small = (unsigned int)(mxGetScalar(prhs[2]));
    n = (unsigned int)(mxGetScalar(prhs[3]));
    k = (unsigned int)(mxGetScalar(prhs[4]));
    XX = (double *)(mxGetPr(prhs[5]));
    YY = (unsigned int *)(mxGetPr(prhs[6]));
    theta = (double *)(mxGetPr(prhs[7]));
    V = (double *)(mxGetPr(prhs[8]));
    S = (double *)(mxGetPr(prhs[9]));
    D_vec = (double *)(mxGetPr(prhs[10]));
    V_in = (double *)(mxGetPr(prhs[11]));
    S_in = (double *)(mxGetPr(prhs[12]));
    D_vec_in = (double *)(mxGetPr(prhs[13]));
    LAMBDA = (double)(mxGetScalar(prhs[14]));
    xi = (double)(mxGetScalar(prhs[15]));
    
    //Outputs
    //Allocate memory and assign output pointer
    plhs[0] = mxCreateDoubleMatrix(DD, 1, mxREAL);
    plhs[1] = mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[2] = mxCreateDoubleMatrix(DD, 1, mxREAL);
    ui = mxGetPr(plhs[0]);
    pzti = mxGetPr(plhs[1]);
    THETA_OUT = mxGetPr(plhs[2]);
    
    //Auxilliary variables
    unsigned int i, j, ii, jj, start, stop, iii;
    Vector g = VectorCreate(DD);
    Vector uti = VectorCreate(DD);
    double logzti;
    double *x1;
    double logai;
    Vector fi_vec = VectorCreate(DD);
    Vector qi = VectorCreate(DD); 
    Vector Nq = VectorCreate(DD); 
    Matrix V_mat = MatrixCreate(DD,k);
    Matrix Vt_mat = MatrixCreate(k,DD);
    Vector tmp_dd = VectorCreate(DD);
    Vector tmp_tab1 = VectorCreate(k);
    Vector tmp_tab2 = VectorCreate(k);
    Matrix S_mat = MatrixCreate(k,k); 
    Matrix AA = MatrixCreate(k,k); 
    Vector ss = VectorCreate(k);
    double norma_kw;
    double c;
    Vector D_vec_vec = VectorCreate(DD); 
    double rat1, rat2;
    double vt, v;
    double suma;
    double tmp;
    double zti;
    
    //Auxilliary variables 
    //Memory allocate
    x1 = (double*)mxCalloc(d_small+1, sizeof(double)); 
    
    VectorSet(qi,0.0);
    VectorSet(Nq,0.0);
    VectorSet((Vector) V_mat, 0.0);
    VectorSet((Vector) Vt_mat, 0.0);
    VectorSet(tmp_dd,0.0);
    VectorSet(tmp_tab1,0.0);
    VectorSet(tmp_tab2,0.0);
    VectorSet(ss,0.0);
    VectorSet(D_vec_vec,0.0);
    
    pzti[0] = 0;
    
    for(i = 0; i <= DD-1; i++)
    {
        ui[i] = 0;
    }
    VectorSet(g,0.0);
    
    for (ii = 0; ii <= k-1; ii++)
    {
       for (jj = 0; jj <= DD-1; jj++)
       { 
          V[jj + ii*DD] = V_in[jj + ii*DD]; 
          V_mat->d2[jj][ii] = V[jj + ii*DD];
          Vt_mat->d2[ii][jj] = V[jj + ii*DD];
       }
    }
    
    for (ii = 0; ii <= k-1; ii++)
    {
       for (jj = 0; jj <= k-1; jj++)
       { 
          S[jj + ii*k] = S_in[jj + ii*k]; 
          S_mat->d2[ii][jj] = S[jj + ii*k];
       }
    }
    
    for (jj = 0; jj <= DD-1; jj++)
    { 
        D_vec[jj] = D_vec_in[jj]; 
        D_vec_vec->d[jj] = D_vec[jj];
    }
    
    for(j = 0; j <= t-1; j++)
    {   
        //mexPrintf("j: %d\n", j);                
        //mexCallMATLAB(0,NULL,0,NULL,"pause");
        //zti = 1e-200;
        logzti = log(1e-200);
        VectorSet(uti,0.0);
        
        for (i = 0; i <= d_small-1; i++)
        {
            x1[i] = XX[j + t*i];
        }
        x1[d_small] = 1;
        
        for (i = 0; i <= n-1; i++)
        {
            //ai = 0;
            logai = 0;
            
            VectorSet(fi_vec,0.0);
            
            start = (d_small+1)*i;
            stop = (d_small+1)*(i+1)-1;
            
            iii = 0;
            for(ii = start; ii <= stop; ii++)
            {                
                fi_vec->d[ii] = x1[iii];
                iii++;
            }   
            
            if(i == YY[j]-1)
            {
                for(ii = 0; ii <= DD-1; ii++)
                { 
                    g->d[ii] = g->d[ii] + fi_vec->d[ii];
                }
            }
                
            for(ii = 0; ii <= DD-1; ii++)
            {                
                logai += theta[ii] * fi_vec->d[ii];
            }
             
            //ai = exp(ai);
            
            VectorSet(qi,0.0);
            VectorAdd(qi,uti);
            VectorSubtract(qi,fi_vec);
            if (logai==logzti)
            {
                VectorScale(qi,0.5);
            }
            else
            {
                vt = logai - logzti;
                v = tanh(0.5*vt);
                v /= (2*vt);
                v = sqrt(v);
                VectorScale(qi,v);
            }     
            
            VectorMove(Nq,qi);
            
            MatrixVectorMultiply(tmp_tab1,Vt_mat,Nq);
            MatrixVectorMultiply(tmp_dd,V_mat,tmp_tab1);
            VectorSubtract(Nq,tmp_dd);
            
            for (ii=0; ii<k; ii++)
            {
                for (jj=0; jj<k; jj++)
                {
                    AA->d2[ii][jj] = S_mat->d2[ii][jj] + tmp_tab1->d[ii]*tmp_tab1->d[jj];
                }
            }
            
            MatrixEigenvectorsSymTony(AA,ss->d);
            
            VectorSet((Vector) S_mat,0.0);
            for (ii=0; ii<k; ii++) 
            {
                S_mat->d2[ii][ii] = ss->d[ii];
            }
            
            MatrixMultiply(Vt_mat,AA,Vt_mat);
            MatrixTranspose(AA);
            MatrixMultiply(V_mat,V_mat,AA);
            
            norma_kw = VectorDot(Nq,Nq);

            double min_v = norma_kw;
            unsigned int min_i = k;
            for (ii=0; ii<k; ii++) 
            {
                if (S_mat->d2[ii][ii] < min_v)
                {
                    min_v = S_mat->d2[ii][ii];
                    min_i = ii;
                }
            }
            
            if (min_i<k) 
            {
                for (jj=0; jj<DD; jj++) 
                {
                    tmp_dd->d[jj] = fabs(V_mat->d2[jj][min_i]);
                }
                suma = VectorSum(tmp_dd);
                c = suma*S_mat->d2[min_i][min_i];
                VectorScale(tmp_dd,c);
                VectorAdd(D_vec_vec,tmp_dd); 
                S_mat->d2[min_i][min_i] = norma_kw;
                norma_kw = 1.0/sqrt(norma_kw);
                VectorScale(Nq,norma_kw);
                for (jj=0; jj<(DD); jj++)
                {
                    V_mat->d2[jj][min_i]  = Nq->d[jj];
                    Vt_mat->d2[min_i][jj] = Nq->d[jj];
                } 
            }
            else
            {
                for (jj=0; jj<DD; jj++) 
                {
                    tmp_dd->d[jj] = fabs(Nq->d[jj]);
                }
                c = VectorSum(tmp_dd);
                VectorScale(tmp_dd,c);
                VectorAdd(D_vec_vec,tmp_dd);
            }
            
            if(logzti > logai)
            {
                rat1 = 1/(1 + exp(logai-logzti));
                rat2 = 1 - rat1;
            }
            else
            {
                rat2 = 1/(exp(logzti-logai) + 1);
                rat1 = 1 - rat2;   
            }                   
            
            VectorScale(uti,rat1);
            VectorMove(tmp_dd,fi_vec);
            VectorScale(tmp_dd,rat2);
            VectorAdd(uti,tmp_dd);

            if(logai < logzti)
            {   
                logzti = logzti + log(1 + exp(logai - logzti));
            }
            else
            {
                logzti = logai + log(1 + exp(logzti - logai));  
            }          
         }
        
        for(jj = 0; jj <= DD-1; jj++)
        {
            uti->d[jj] += LAMBDA * theta[jj];
            ui[jj] += uti->d[jj];  
        }
        
        tmp = 0;
        
        for(ii = 0; ii <= DD-1; ii++)
        {
            tmp += theta[ii] * theta[ii];
        }
        
        //zti -= 1e-200;
        //zti = log(zti);
        logzti = logzti + log(1.0-exp(log(1e-200)-logzti));
        zti = logzti + 0.5 * LAMBDA * tmp;
        pzti[0] += zti;
    } 
            
    for(jj = 0; jj <= DD-1; jj++)
    {
        ui[jj] -= g->d[jj];  
    }  
    
    tmp = 0;
    for(jj = 0; jj <= DD-1; jj++)
    {
        tmp += theta[jj] * g->d[jj];
    }
    pzti[0] -= tmp;
    
    //(A + BC)^-1 = A^-1 - A^-1 B (I + C A-1 B)^-1 C A^-1.
    // Invert lowrank&diagonal covariance prior to qi multiply
    //                       ( A + B   C    )
    // Use Woodbury above on ( D + V'*(S*V) )^{-1}
    // where A->D B->V' and C->SV
    // so inv = D^-1 - D^-1 V' (I+ (S*V) D^-1 V' )^-1 S*V  D^-1

    for (jj=0; jj<DD; jj++) 
    {
        tmp_dd->d[jj] = ui[jj]/D_vec_vec->d[jj];
    }
    for (ii=0; ii<k; ii++) 
    {
        for (jj=0; jj<DD; jj++) 
        {
            Vt_mat->d2[ii][jj] /= D_vec_vec->d[jj];
        }
    }
    MatrixMultiply(AA,Vt_mat,V_mat);
    MatrixMultiply(AA,S_mat,AA);
    for (ii=0; ii<k; ii++)
    {
        AA->d2[ii][ii] += 1.0;
    }
    MatrixInvert(AA);
    MatrixMultiply(AA,AA,S_mat);
    MatrixVectorMultiply(tmp_tab1,Vt_mat,tmp_dd);
    MatrixVectorMultiply(tmp_tab2,AA,tmp_tab1);
    for (ii=0; ii<k; ii++) 
    {
        for (jj=0; jj<DD; jj++) 
        {
            Vt_mat->d2[ii][jj] = V_mat->d2[jj][ii];
        }
    }
    MatrixVectorMultiply(Nq,V_mat,tmp_tab2);
    for (jj=0; jj<DD; jj++) 
    {
        Nq->d[jj] = ui[jj]/D_vec_vec->d[jj] - Nq->d[jj]/D_vec_vec->d[jj];
    }
    for (jj=0; jj<DD; jj++) 
    {
        THETA_OUT[jj] = theta[jj] - xi*Nq->d[jj];
    }
      
    //Delete
    VectorFree(g);
    VectorFree(uti);
    VectorFree(fi_vec);
    VectorFree(qi);
    VectorFree(Nq);
    MatrixFree(V_mat);
    MatrixFree(Vt_mat);
    VectorFree(tmp_dd);
    VectorFree(tmp_tab1);
    VectorFree(tmp_tab2);
    MatrixFree(S_mat);
    MatrixFree(AA);
    VectorFree(ss);
    VectorFree(D_vec_vec);
}    
    
    



