% Code by Anna Choromanska and Tony Jebara
% Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

function [zi,ui] = logZ(THETA)
global DD;
global d_small;
global t;
global n;
global cc;
global XX;
global YY;
global LAMBDA;
global logZtab;
global flag_alg_on;

if(flag_alg_on == 1)
    save('RESULT_LBFGS.mat','logZtab');
elseif(flag_alg_on == 2)
    save('RESULT_SD.mat','logZtab');
elseif(flag_alg_on == 3)
    save('RESULT_CG.mat','logZtab');
end

zi = 0.0;
ui = zeros(DD,1);
GG=zeros(1,DD);
for j=1:t
    x1=[XX(j,:) 1];
    f=sparse(cc,DD);
    for c=1:cc
        f(c,(d_small+1)*(c-1)+1:(d_small+1)*c)=x1;
        if (c==YY(j))
            GG = GG+f(c,:);
        end
    end
    n=size(f,1);
    logzti = log(1e-200);
    uti = zeros(DD,1);
    for i=1:n
        fi = f(i,:)';
        logai = dot(THETA,fi);
        uti = uti + exp(logai)*fi;
        if(logai < logzti)
            logzti = logzti + log(1 + exp(logai - logzti));
        else
            logzti = logai + log(1 + exp(logzti - logai));
        end
    end
    uti = uti/exp(logzti);
    uti = uti + LAMBDA*THETA;
    logzti = logzti + log(1.0-exp(log(1e-200)-logzti));
    zti=logzti+0.5*LAMBDA*dot(THETA,THETA);
    ui  = ui+uti;
    zi  = zi+zti;
end

ui = ui - GG';
zi = zi - dot(THETA,GG')';

if((zi > 1e300) || (isnan(zi)))
    zi = 1e300;
elseif(zi < -1e300)
    zi = -1e300;
end

logZtab = [logZtab;zi,cputime];

disp(size(logZtab,1));
disp(zi);

% COMMENTED CODE BELOW IS EQUIVALENT TO THE CODE ABOVE, HOWEVER LESS NUMERICALLY
% STABLE
% zi = 0.0;
% ui = zeros(DD,1);
% GG=zeros(1,DD);
% for j=1:t
%     %disp(j);
%     x1=[XX(j,:) 1];
%     f=sparse(cc,DD);
%     for c=1:cc
%         f(c,(d_small+1)*(c-1)+1:(d_small+1)*c)=x1;
%         if (c==YY(j))
%             GG = GG+f(c,:);
%         end
%     end
%     n=size(f,1);
%     zti = 1e-200;
%     uti = zeros(DD,1);
%     for i=1:n
%         fi = f(i,:)';
%         dot(THETA,fi);
%         ai = exp(dot(THETA,fi));
%         uti = uti + ai*fi;
%         zti = zti + ai;
%     end
%     uti = uti/zti;
%     uti = uti + LAMBDA*THETA;
%     zti = zti-1e-200;
%     zti = log(zti) + 0.5*LAMBDA*dot(THETA,THETA);
%     ui  = ui+uti;
%     zi  = zi+zti;
% end
% 
% ui = ui - GG';
% zi = zi - dot(THETA,GG')';
% 
% if((zi > 1e300) || (isnan(zi)))
%     zi = 1e300;
% elseif(zi < -1e300)
%     zi = -1e300;
% end
% 
% logZtab = [logZtab;zi,cputime];
% 
% disp(size(logZtab,1));
% disp(zi);