% Code by Anna Choromanska and Tony Jebara
% Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

function [RESULT_LBFGS,RESULT_SD,RESULT_CG,RESULT_BOUND_Matlab_v1,RESULT_BOUND_Mex_v2] = realbound(XX,YY,lamb,maxiter,TolX_val,TolFun_val,file_now,ll,iteration)
% GLOBALS
global LAMBDA;
global logZtab;
global DD;
global d_small;
global t;
global n;
global cc;
global XX;
global YY;
global flag_alg_on;
global flag_theta_gen;
global flag_LBFGS;
global flag_SD;
global flag_CG;
global flag_BOUND_Matlab_v1;
global flag_BOUND_Mex_v2;

% OUTPUT VARIABLES INITIALIZATION
RESULT_LBFGS = [1,1];
RESULT_SD = [1,1];
RESULT_CG = [1,1];
RESULT_BOUND_Matlab_v1 = [1,1];
RESULT_BOUND_Mex_v2 = [1,1];
 
% VARIABLES
LAMBDA=lamb;
[t,d]=size(XX);
d_small = d;
cc=length(unique(YY));
DD=(d+1)*cc;
disp('[t,DD,cc]');
disp([t,DD,cc]);
k = ceil(log10(DD));
theta1 = zeros(DD,1);
if(flag_theta_gen == 1)
    theta1 = (rand(DD,1)-0.5)./10^10;
    s=sprintf('LogReg_theta_initialization%d.%d.%d.mat',file_now,ll,iteration); save(s,'theta1');  
else
    s=sprintf('LogReg_theta_initialization%d.%d.%d.mat',file_now,ll,iteration); load(s);   
end

s=sprintf('LogReg%d.%d.%d.mat',file_now,ll,iteration); 
if(exist(s))
    load(s);
end

% LBFGS
%--------------------------------------------------------------------------
if(flag_LBFGS == 1)
    disp('LBFGS');
    flag_alg_on = 1;
    RESULT_LBFGS = [];
    logZtab = [];
    cpustart = cputime;
    options = struct('HessUpdate','lbfgs','StoreN',k,'GradObj','on','MaxFunEvals',maxiter,'TolX',TolX_val,'TolFun',TolFun_val);
    [X,FVAL,EXITFLAG,OUTPUT]=fminlbfgs(@logZ,theta1,options);
    STOPPER = FVAL+1e-4; % log(0.999*exp(FVAL)+0.001*exp(logZ(0*X)));
    s=sprintf('LogReg_STOPPER%d.%d.%d.mat',file_now,ll,iteration); save(s,'STOPPER');
    logZtab(:,2) = logZtab(:,2) - cpustart;
    RESULT_LBFGS = logZtab;
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_LBFGS,1)
        if(RESULT_LBFGS(i,1) > RESULT_LBFGS(i-1,1))
            RESULT_LBFGS(i,1) = RESULT_LBFGS(i-1,1);
        end
    end
    %-------------------------------------------------------------------- 
end
%--------------------------------------------------------------------------

% Steepest descent
%--------------------------------------------------------------------------
if(flag_SD == 1) 
    disp('Steepest descent');
    flag_alg_on = 2;
    RESULT_SD = [];
    logZtab = [];
    cpustart = cputime;
    options = struct('LargeScale','off','HessUpdate','steepdesc','GradObj','on','MaxFunEvals',maxiter,'TolX',TolX_val,'TolFun',TolFun_val);%,'Display','iter');
    [X,FVAL,EXITFLAG,OUTPUT]=fminlbfgs(@logZ,theta1,options);
    logZtab(:,2) = logZtab(:,2) - cpustart;
    RESULT_SD = logZtab;
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_SD,1)
        if(RESULT_SD(i,1) > RESULT_SD(i-1,1))
            RESULT_SD(i,1) = RESULT_SD(i-1,1);
        end
    end
    %--------------------------------------------------------------------    
end
%--------------------------------------------------------------------------

% Conjugate gradient
%--------------------------------------------------------------------------
if(flag_CG == 1)
    disp('Conjugate gradient');
    flag_alg_on = 3;
    RESULT_CG = [];
    logZtab = [];
    cpustart = cputime;
    [x fx c] = minimize(theta1, 'logZ', -maxiter); %max of maxiter linesearches, if I write - maxiter I have maximum of maxiter function evauations
    logZtab(:,2) = logZtab(:,2) - cpustart;
    RESULT_CG = logZtab;
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_CG,1)
        if(RESULT_CG(i,1) > RESULT_CG(i-1,1))
            RESULT_CG(i,1) = RESULT_CG(i-1,1);
        end
    end
    %--------------------------------------------------------------------    
end
%--------------------------------------------------------------------------

%Bound with sigma inversion xi = 1.0 Matlab implementation (equivalent to Mex implementation however slower)
%--------------------------------------------------------------------------
%k = DD;
if(flag_BOUND_Matlab_v1 == 1)
    xi = 1.0;
    disp('Bound with sigma inversion xi = 1.0 Matlab v.1');
    RESULT_BOUND_Matlab_v1 = [];
    cpustart = cputime;
    THETA = theta1;
    iter = 0;
    flag_iter = 1;
    flag_TolX = 1;
    flag_TolFun = 1;
    while(flag_iter && flag_TolX && flag_TolFun)
        zi = 0.0;
        ui=zeros(DD,1);        
        %------------------------------------------------------------------
        V = zeros(DD,k);
        V_tmp = rand(k,k);
        [Q,R] = qr(V_tmp);
        V(1:k,1:k) = Q; 
        D_vec = ones(DD,1).*(t*LAMBDA);
        S = zeros(k,k);
        %------------------------------------------------------------------
        GG=zeros(1,DD);
        %S_dod = eye(DD,DD)*t*LAMBDA;
        for j=1:t
            %St_dod = 1e-200*eye(DD,DD);
            %zti = 1e-200;
            logzti = log(1e-200);
            uti = zeros(DD,1);
            x1=[XX(j,:) 1];
            f=sparse(cc,DD);
            for c=1:cc
                f(c,(d_small+1)*(c-1)+1:(d_small+1)*c)=x1;
                if (c==YY(j))
                    GG = GG+f(c,:);
                end
            end
            n=size(f,1);
            for i=1:n
                fi = f(i,:)';
                %ai = exp(dot(THETA,fi));
                logai = dot(THETA,fi);
                %aizi=ai/zti;
                %if ((aizi)==1)
                if(logai == logzti)
                    qi = 0.5*(uti-fi);
                else
                    %qi = sqrt( tanh(0.5*(log(ai) - log(zti)))/(2*(log(ai) - log(zti))) )*(uti-fi);
                    qi = sqrt( tanh(0.5*(logai - logzti))/(2*(logai - logzti)) )*(uti-fi);
                end
                %St_dod = St_dod + qi*qi';
                %------------------------------------------------------------------
                Nq = qi;
                tmp_tab1 = V'*Nq;
                tmp_dd = V*tmp_tab1; 
                Nq = Nq - tmp_dd;
                S = S + tmp_tab1*tmp_tab1';
                [Vq,Sq] = eig(S);
                S = Sq;
                V = V*Vq;               
                norma_kw = norm(Nq)^2;
                S_tmp = [diag(S); norma_kw];
                [value,index] = min(S_tmp);
                if(index <= k)
                    c = S(index,index)*sum(abs(V(:,index)),1);
                    D_vec = D_vec + c*(abs(V(:,index)));
                    S(index,index) = norma_kw;        
                    Nq = Nq/sqrt(norma_kw);
                    V(:,index) = Nq;
                else
                    c = sum(abs(Nq),1);
                    D_vec = D_vec + c*abs(Nq);
                end
                %------------------------------------------------------------------
                %uti = zti/(zti+ai)*uti + ai/(zti+ai)*fi;
                if(logzti>logai)
                    rat1 = 1.0/(1.0+exp(logai-logzti));
                    rat2 = 1.0-rat1;
                else
                    rat2 = 1.0/(exp(logzti-logai)+1.0);
                    rat1 = 1.0-rat2;
                end   
                uti = rat1*uti + rat2*fi;
                %zti = zti + ai;
                if(logai<logzti) 
                    logzti = logzti + log(1.0+exp(logai-logzti));
                else
                    logzti = logai + log(1.0+exp(logzti-logai));
                end
            end
            uti=uti+LAMBDA*THETA;
            %zti=zti-1e-200;
            logzti = logzti + log(1.0-exp(log(1e-200)-logzti));
            %zti=log(zti)+0.5*LAMBDA*dot(THETA,THETA);
            zti=logzti+0.5*LAMBDA*dot(THETA,THETA);
            ui = ui+uti;
            zi = zi+zti;
            %S_dod = S_dod + St_dod;
        end
        ui = ui-GG';
        zi = zi-dot(THETA,GG');
        %S_dod = (1/xi)*S_dod;
        %THETA = THETA-pinv(S_dod)*ui;  
        tmp_dd = ui./D_vec;
        Vt = V';
        for ii=1:k
            Vt(ii,:) = Vt(ii,:)./(D_vec');
        end
        AA = Vt*V;
        AA = S*AA;
        AA = AA + eye(k,k);
        AA = pinv(AA);
        AA = AA*S;
        tmp_tab1 = Vt*tmp_dd;
        tmp_tab2 = AA*tmp_tab1;
        Vt = V';
        Nq = V*tmp_tab2;
        Nq = ui./D_vec - Nq./D_vec;
        THETA = THETA - xi*Nq;
        %Alternative update on THETA, however more computationally
        %expensive
        %THETA = THETA-xi*pinv(V*S*V' + diag(D_vec))*ui;
        RESULT_BOUND_Matlab_v1 = [RESULT_BOUND_Matlab_v1;zi,cputime-cpustart];
        save('RESULT_BOUND_Matlab_v1.mat','RESULT_BOUND_Matlab_v1');
        iter = iter + 1;
        disp(zi);
        disp(iter);
        %-----------------------Stopping criteria------------------------------
        if(iter > maxiter)
            disp('Optimization completed because the maximum number of iterations was reached');
            flag_iter = 0;
        end
        if(iter > 3)
            if((norm(RESULT_BOUND_Matlab_v1(iter)- RESULT_BOUND_Matlab_v1(iter-1)) < TolFun_val) && (norm(RESULT_BOUND_Matlab_v1(iter-1)- RESULT_BOUND_Matlab_v1(iter-2)) < TolFun_val) && (norm(RESULT_BOUND_Matlab_v1(iter-2)- RESULT_BOUND_Matlab_v1(iter-3)) < TolFun_val))
                disp('Optimization completed because the change of function value is less than the size of the function tolerance');
                flag_TolFun = 0;
            end
        end
    %     if(norm(THETA - theta_tmp) < TolX_val)
    %         disp('Optimization completed because the change of parameter vector is less than the size of the parameter tolerance');
    %         flag_TolX = 0;
    %     end
        %----------------------------------------------------------------------
    end
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_BOUND_Matlab_v1,1)
        if(RESULT_BOUND_Matlab_v1(i,1) > RESULT_BOUND_Matlab_v1(i-1,1))
            RESULT_BOUND_Matlab_v1(i,1) = RESULT_BOUND_Matlab_v1(i-1,1);
        end
    end
    %--------------------------------------------------------------------
end
%--------------------------------------------------------------------------

% Bound with sigma inversion xi = 1.0 Mex implementation (equivalent to Matlab implementation however faster)
%--------------------------------------------------------------------------
if(flag_BOUND_Mex_v2 == 1)
    xi = 1.0;
    disp('Bound with sigma inversion xi = 1.0 Mex v.2');
    RESULT_BOUND_Mex_v2 = [];
    %cpustart = cputime;
    THETA = theta1;
    iter = 0;
    flag_iter = 1;
    flag_TolX = 1;
    flag_TolFun = 1;
    %------------------------------------------------------------------       
    V_in = zeros(DD,k);
    V_tmp = rand(k,k);
    [Q,R] = qr(V_tmp);
    V_in(1:k,1:k) = Q; 
    D_vec_in = ones(DD,1).*(t*LAMBDA);
    S_in = zeros(k,k);
   
    V = zeros(DD,k);
    V_tmp = rand(k,k);
    [Q,R] = qr(V_tmp);
    V(1:k,1:k) = Q; 
    D_vec = ones(DD,1).*(t*LAMBDA);
    S = zeros(k,k);
    
    n = cc;
    %------------------------------------------------------------------
    cpustart = cputime;
    while(flag_iter && flag_TolX && flag_TolFun)  
        %------------------------------------------------------------------
        save('RESULT_BOUND_Mex_v2.mat','RESULT_BOUND_Mex_v2');
        [ui,zi,THETA] = fast_k_bound(t,DD,d_small,n,k,XX,uint32(YY),THETA,V,S,D_vec,V_in,S_in,D_vec_in,LAMBDA,xi);
        RESULT_BOUND_Mex_v2 = [RESULT_BOUND_Mex_v2;zi,cputime-cpustart];
        save('RESULT_BOUND_Mex_v2.mat','RESULT_BOUND_Mex_v2');
        iter = iter + 1;
        disp(zi);
        disp(iter);
        %-----------------------Stopping criteria------------------------------
        if(iter > maxiter)
            disp('Optimization completed because the maximum number of iterations was reached');
            flag_iter = 0;
        end
        if(iter > 3)
            if((norm(RESULT_BOUND_Mex_v2(iter)- RESULT_BOUND_Mex_v2(iter-1)) < TolFun_val) && (norm(RESULT_BOUND_Mex_v2(iter-1)- RESULT_BOUND_Mex_v2(iter-2)) < TolFun_val) && (norm(RESULT_BOUND_Mex_v2(iter-2)- RESULT_BOUND_Mex_v2(iter-3)) < TolFun_val))
                disp('Optimization completed because the change of function value is less than the size of the function tolerance');
                flag_TolFun = 0;
            end
        end
    %     if(norm(THETA - theta_tmp) < TolX_val)
    %         disp('Optimization completed because the change of parameter vector is less than the size of the parameter tolerance');
    %         flag_TolX = 0;
    %     end
        %----------------------------------------------------------------------
    end
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_BOUND_Mex_v2,1)
        if(RESULT_BOUND_Mex_v2(i,1) > RESULT_BOUND_Mex_v2(i-1,1))
            RESULT_BOUND_Mex_v2(i,1) = RESULT_BOUND_Mex_v2(i-1,1);
        end
    end
    %--------------------------------------------------------------------
end
%--------------------------------------------------------------------------
