% Code by Anna Choromanska and Tony Jebara
% Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

%----------------------------------------------------------------------    
%time/iterations - the time/number of iterations in which the
%algorithms converge to the same solution as LBFGS if they can approach
%it or the time/number of iterations in which the
%algorithms converge (if they can not approach LBFGS solution)

%time_p/iterations_p - the time/number of iterations in which the
%algorithms converge to the same solution as (solution of LBFGS) - p*(solution of LBFGS - starting point)
%if they can approach it or the time/number of iterations in which the
%algorithms converge (if they can not approach (solution of LBFGS) -
%p*(solution of LBFGS - starting point))

%value/value_t - the value which the algorithms approach in the same number of
%iterations/time as the number of iterations/time of convergence of LBFGS if they
%converge in more iterations/time than LBFGS or the value to which the
%algorithms converge (if they converge in less iterations/time than LBFGS)

%value_p/value_t_p - the value which the algorithms approach in the same number of
%iterations/time as the number of iterations/time of convergence of LBFGS
%to solution (solution of LBFGS) - p*(solution of LBFGS - starting point) if they
%converge in more iterations/time than the number of iterations/time of convergence of 
%LBFGS to solution (solution of LBFGS) - p*(solution of LBFGS - starting point)
%or the value to which the algorithms converge (if they converge in less iterations/time 
%than the number of iterations/time of convergence of LBFGS to solution
%(solution of LBFGS) - p*(solution of LBFGS - starting point))

%value_conv the value to which the algorithms converge

%percent - percent of iterations the algorithms do better than LBFGS
%percent_t - percent of time the algorithms do better than LBFGS
%----------------------------------------------------------------------
function [FINAL_TAB] = analysis(maxiter,p,file_now,iteration)
    global flag_BFGS;
    global flag_SD;
    global flag_CG;
    global flag_NEWTON_v1;
    global flag_BOUND_xi1_Mex_v2;
 
    %Loading results
    s=sprintf('Latent_STOPPER%d.%d.mat',file_now,iteration); load(s);
    s=sprintf('Latent%d.%d.mat',file_now,iteration); load(s);

    FINAL_TAB = [];

    i = 1;
    RESULT_CELL{i} = RESULT_BFGS;
    
    if(flag_SD == 1)
        i = i + 1;
        RESULT_CELL{i} = RESULT_SD;
    end
    if(flag_CG == 1)
        i = i + 1;
        RESULT_CELL{i} = RESULT_CG;
    end
    if(flag_NEWTON_v1 == 1)
        i = i + 1;
        RESULT_CELL{i} = RESULT_NEWTON_v1;
    end
    if(flag_BOUND_xi1_Mex_v2 == 1)
        i = i + 1;
        RESULT_CELL{i} = RESULT_BOUND_xi1_Mex_v2;
    end

    %----------------------------------------------------------------------
    %Cutting the tables to maxiter iterations
    for i = 1:1:size(RESULT_CELL,2) - 1
        RESULT_CELL{i} = RESULT_CELL{i}(1:min(maxiter,size(RESULT_CELL{i},1)),:);
    end
    %----------------------------------------------------------------------
    
    RESULT_BFGS = [];
    RESULT_BFGS = RESULT_CELL{1};
%     if(RESULT_BFGS(size(RESULT_BFGS,1),1)) > STOPPER
%         STOPPER = RESULT_BFGS(size(RESULT_BFGS,1),1);
%     end
    
    %BFGS
    index = [];
    index = find(RESULT_BFGS(:,1) <= STOPPER);
    time = RESULT_BFGS(index(1),2);
    value = RESULT_BFGS(index(1),1);
    iterations = index(1);
    iterations_BFGS = iterations;
    %----------------------------------------------------------------------
    value_t = value;
    %----------------------------------------------------------------------
    value_conv = value;
    %---------------------------------------------------------------------- 
    percent = 0;
    %---------------------------------------------------------------------- 
    percent_t = 0;
    %---------------------------------------------------------------------- 
    index = [];
    index = find(RESULT_BFGS(:,1) <= (STOPPER+p*abs(STOPPER-RESULT_BFGS(1,1))));
    time_p = RESULT_BFGS(index(1),2);
    iterations_p = index(1);
    iterations_p_BFGS = iterations_p;
    time_p_BFGS = time_p;
    value_p = RESULT_BFGS(index(1),1);
    %----------------------------------------------------------------------
    value_t_p = value_p;
    %---------------------------------------------------------------------- 
    if(flag_BFGS == 1)
        FINAL_TAB = [FINAL_TAB;value,time,iterations,value_t,percent,percent_t,value_conv,time_p,iterations_p,value_p,value_t_p];
    end
    time_BFGS = time;
    
    for i = 2:1:size(RESULT_CELL,2) 
        RESULT_TAB = [];
        RESULT_TAB = RESULT_CELL{i};
        
        index = [];
        index = find(RESULT_TAB(:,1) <= STOPPER);
        if((size(index,1) ~= 0) && (size(index,2) ~= 0))
            time = RESULT_TAB(index(1),2);
            iterations = index(1);
        else
            time = RESULT_TAB(size(RESULT_TAB,1),2);
            iterations = size(RESULT_TAB,1);
        end
        value = RESULT_TAB(min(iterations_BFGS,size(RESULT_TAB,1)),1);
        %----------------------------------------------------------------------
        index = [];
        index = find(RESULT_TAB(:,2) >= time_BFGS);
        if((size(index,1) ~= 0) && (size(index,2) ~= 0))
            value_t = RESULT_TAB(index(1),1);
        else      
            value_t = RESULT_TAB(size(RESULT_TAB,1),1);
        end
        %----------------------------------------------------------------------
        value_conv = RESULT_TAB(size(RESULT_TAB,1),1);
        %----------------------------------------------------------------------
        percent = 0;
        for i = 1:1:min(size(RESULT_BFGS,1),size(RESULT_TAB,1))
            if((-1)*RESULT_TAB(i,1) > (-1)*RESULT_BFGS(i,1))
                percent = percent + 1;
            end
        end
        percent = (percent/min(size(RESULT_BFGS,1),size(RESULT_TAB,1)))*100;
        %---------------------------------------------------------------------- 
        percent_t = 0;
        m1 = RESULT_BFGS(size(RESULT_BFGS,1),2);
        m2 = RESULT_TAB(size(RESULT_TAB,1),2);
        if m1 > m2
            index = [];
            index = find(RESULT_BFGS(:,2) <= m2);
            stop_BFGS = size(index,1);
            stop_alg = size(RESULT_TAB,1);
        else
            index = [];
            index = find(RESULT_TAB(:,2) <= m1);
            stop_alg = size(index,1);
            stop_BFGS = size(RESULT_BFGS,1);
        end
        if(stop_BFGS >= stop_alg)
            for i = 1:1:stop_BFGS
                index1 = [];
                index2 = [];
                index1 = find(RESULT_TAB(:,2) <= RESULT_BFGS(i,2));
                index2 = find(RESULT_TAB(:,2) >= RESULT_BFGS(i,2));
                if((size(index1,1) == 0) || (size(index1,2) == 0))
                    comp = RESULT_TAB(index2(1),1);
                elseif((size(index2,1) == 0) || (size(index2,2) == 0))
                    comp = RESULT_TAB(index1(size(index1,1)),1);
                else
                    if(abs(RESULT_TAB(index1(size(index1,1)),2) - RESULT_BFGS(i,2)) >= abs(RESULT_TAB(index2(1),2) - RESULT_BFGS(i,2)))
                        comp = RESULT_TAB(index2(1),1);
                    else
                        comp = RESULT_TAB(index1(size(index1,1)),1);
                    end
                end
                if((-1)*comp > (-1)*RESULT_BFGS(i,1))
                    percent_t = percent_t + 1;
                end
            end
            percent_t = (percent_t/stop_BFGS)*100;
        else
            for i = 1:1:stop_alg
                index1 = [];
                index2 = [];
                index1 = find(RESULT_BFGS(:,2) <= RESULT_TAB(i,2));
                index2 = find(RESULT_BFGS(:,2) >= RESULT_TAB(i,2));
                if((size(index1,1) == 0) || (size(index1,2) == 0))
                    comp = RESULT_BFGS(index2(1),1);
                elseif((size(index2,1) == 0) || (size(index2,2) == 0))
                    comp = RESULT_BFGS(index1(size(index1,1)),1);
                else
                    if(abs(RESULT_BFGS(index1(size(index1,1)),2) - RESULT_TAB(i,2)) >= abs(RESULT_BFGS(index2(1),2) - RESULT_TAB(i,2)))
                        comp = RESULT_BFGS(index2(1),1);
                    else
                        comp = RESULT_BFGS(index1(size(index1,1)),1);
                    end
                end
                if((-1)*RESULT_TAB(i,1) > (-1)*comp)
                    percent_t = percent_t + 1;
                end
            end
            percent_t = (percent_t/stop_alg)*100;
        end
        %---------------------------------------------------------------------- 
        index = [];
        index = find(RESULT_TAB(:,1) <= (STOPPER+p*abs(STOPPER-RESULT_BFGS(1,1))));
        if((size(index,1) ~= 0) && (size(index,2) ~= 0))
            time_p = RESULT_TAB(index(1),2);
            iterations_p = index(1);
        else
            time_p = RESULT_TAB(size(RESULT_TAB,1),2);
            iterations_p = size(RESULT_TAB,1);
        end
        %----------------------------------------------------------------------
        value_p = RESULT_TAB(min(iterations_p_BFGS,size(RESULT_TAB,1)),1);
        %----------------------------------------------------------------------
        index = [];
        index = find(RESULT_TAB(:,2) >= time_p_BFGS);
        if((size(index,1) ~= 0) && (size(index,2) ~= 0))
            value_t_p = RESULT_TAB(index(1),1);
        else      
            value_t_p = RESULT_TAB(size(RESULT_TAB,1),1);
        end
        %----------------------------------------------------------------------
        FINAL_TAB = [FINAL_TAB;value,time,iterations,value_t,percent,percent_t,value_conv,time_p,iterations_p,value_p,value_t_p];
    end
end
