// Code by Anna Choromanska and Tony Jebara
// Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mex.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    //inputs
    unsigned int T,M,Y,D;
    double *F;        
    double *theta;
    double *S;
    double *YY;
    double LAMBDA;
    
	//calculated
    
	//variables for loops
	unsigned int i,j, addr, tmpi;
    unsigned int t, m, y, d;
    double ai, v, vt, vx;
    
    double zt;
    double *tmp_tab;
    double *ut;
    double *fi;
    double *nf;
    double *eta;
    
    double den, num, tmp;
    
	//output
	double *St;
    double *S_out;
    double *pden;
    double *pnum;
    double *u;
    
    //get inputs
    T = (unsigned int)(mxGetScalar(prhs[0]));
    M = (unsigned int)(mxGetScalar(prhs[1]));
    Y = (unsigned int)(mxGetScalar(prhs[2]));
    D = (unsigned int)(mxGetScalar(prhs[3]));
    F = (double *)(mxGetPr(prhs[4]));
    theta = (double *)(mxGetPr(prhs[5]));
    S = (double *)(mxGetPr(prhs[6]));
    YY = (double *)(mxGetPr(prhs[7]));
    LAMBDA = (double)(mxGetScalar(prhs[8]));
    
    //Outputs
    //Allocate memory and assign output pointer
    plhs[0] = mxCreateDoubleMatrix(D, D, mxREAL);
    plhs[1] = mxCreateDoubleMatrix(D, D, mxREAL);
    plhs[2] = mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[3] = mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[4] = mxCreateDoubleMatrix(D, 1, mxREAL);
    St = mxGetPr(plhs[0]);	
    S_out = mxGetPr(plhs[1]);	
    pden = mxGetPr(plhs[2]);
    pnum = mxGetPr(plhs[3]);
    u = mxGetPr(plhs[4]);
    
    //Memory allocate
    tmp_tab = (double*)mxCalloc(D, sizeof(double));
    ut = (double*)mxCalloc(D, sizeof(double));
    fi = (double*)mxCalloc(D, sizeof(double));
    nf = (double*)mxCalloc(D, sizeof(double));
    eta = (double*)mxCalloc(M, sizeof(double));
    
    for(i = 0; i <= D-1; i++)
    {
        u[i] = 0;
        
        for(j = 0; j <= D-1; j++)
        {
            S_out[i+D*j] = S[i+D*j];
        }    
    }
    
    den = 0;
    num = 0;
    
    for(t = 0; t <= T-1; t++)
    {
        zt = 1e-200;
        
        for(i = 0; i <= D-1; i++)
        {
            for(j = 0; j <= D-1; j++)
            {
                St[i + j*D] = 0;
            }
            
            St[i + i*D] = zt;
            ut[i] = 0;
        }
        
        for(m = 0; m <= M-1; m++)
        {
            for(y = 0; y <= Y-1; y++)
            {
                addr = D*(t+T*(y+Y*m));
                
                ai = 0;
                
                for(j = 0; j <= D-1; j++)
                {
                    fi[j] = F[addr];
                    addr++;
                    
                    ai += theta[j] * fi[j];
                }
                
                ai = exp(ai);
                
                if((ai/zt) == 1)
                {           
                    for (i = 0; i <= D-1; i++)
                    {
                        tmp_tab[i] = ut[i] - fi[i];
                    }
                    
                    for(i = 0; i <= D-1; i++)
                    {
                        for(j = 0; j <= D-1; j++)
                        {   
                            addr = j + i * D;
                            v = tmp_tab[i]*tmp_tab[j];
                            v *= 0.25;
                            St[addr] = St[addr] + v;
                        }
                    }
                }
                else
                {
                    vt = log(ai/zt);
                    v = tanh(0.5*vt);
                    v /= (2*vt);

                    for (i = 0; i <= D-1; i++)
                    {
                        tmp_tab[i] = ut[i] - fi[i];
                    }
                    for(i = 0; i <= D-1; i++)
                    {
                        for(j = 0; j <= D-1; j++)
                        {   
                            addr = j + i * D;                
                            vx = tmp_tab[i]*tmp_tab[j];
                            vx *= v;
                            St[addr] = St[addr] + vx;
                        }
                    }
                }
                
                for(i = 0; i <= D-1; i++)
                {
                    ut[i] *= zt/(zt+ai);
                    ut[i] += (ai/(zt+ai))*fi[i];
                }
                zt += ai;                        
            }
        }
        
        for(i = 0; i <= D-1; i++)
        {
            nf[i] = 0;
        }
        
        for(m = 0; m <= M-1; m++)
        {
            eta[m] = 0;
            
            tmpi = (unsigned int)YY[t] - 1;
            addr = (D*(t+T*(tmpi+Y*m)));
                
            for(j = 0; j <= D-1; j++)
            {
                fi[j] = F[addr];
                addr++;
                
                eta[m] += theta[j] * fi[j];
            }    
            
            eta[m] = exp(eta[m]);
            
            for(j = 0; j <= D-1; j++)
            {
                nf[j] += fi[j] * eta[m];
            }
        }
        
        tmp = 0;
        
        for(j = 0; j <= M-1; j++)
        {
            tmp += eta[j];
        } 
        
        for(j = 0; j <= D-1; j++)
        {
            nf[j] /= tmp;               
        }
        
        for(i = 0; i <= D-1; i++)
        {
            u[i] += ut[i] - nf[i];
            u[i] += LAMBDA * theta[i];

            for(j = 0; j <= D-1; j++)
            {
                S_out[i+D*j] += St[i+D*j];
            }    
        }
        
        num += log(tmp);
        
        tmp = 0;
        
        for(i = 0; i <= D-1; i++)
        {
            tmp += theta[i] * theta[i];
        }
        
        num -= 0.5 * LAMBDA * tmp;
        den += log(zt);
    }
    
    pnum[0] = num;
    pden[0] = den;    
}
