% Code by Anna Choromanska and Tony Jebara
% Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

function [RESULT_BFGS,RESULT_SD,RESULT_CG,RESULT_NEWTON_v1,RESULT_BOUND_xi1_Mex_v2] = latentlikelihood(XX,YY,YY_all,lambda,M,maxiter,TolX_val,TolFun_val, file_now, iteration)
%
% function [score,SCOREG]=latentlikelihood(XX,YY,xi,M)
%

global LAMBDA;
global T;
global M;
global Y;
global F;
global YY;
global flag_alg_on;
global flag_theta_gen;
global dims;
global latenttab;
global flag_BFGS;
global flag_SD;
global flag_CG;
global flag_NEWTON_v1;
global flag_BOUND_xi1_Mex_v2;

% OUTPUT VARIABLES INITIALIZATION
RESULT_BFGS = [1,1,1];
RESULT_SD = [1,1,1];
RESULT_CG = [1,1,1];
RESULT_NEWTON_v1 = [1,1,1];
RESULT_BOUND_xi1_Mex_v2 = [1,1,1];

% VARIABLES
LAMBDA=lambda;
T=size(XX,1);
D=size(XX,2);
Y=length(unique(YY_all));
dims=(D+1)*M*Y;
scaling_factor = 1e-5; 

F = cell(T,Y,M);
for t=1:T
    for y=1:Y
        for m=1:M
            F{t,y,m} = zeros((D+1)*M*Y,1);
        end
    end
end

for t=1:T
    for y=1:Y
        for m=1:M
            f=zeros((D+1)*M*Y,1);
            f( ((y-1)*M*(D+1) + (m-1)*(D+1))+1: (((y-1)*M*(D+1) + (m-1)*(D+1))+D+1) )=[XX(t,:)'; 1];
            F{t,y,m} = f;
        end
    end
end

[FM] = ArrayToMatrixconversion(F);

theta1 = zeros(dims,1);
if(flag_theta_gen == 1)
    theta1 = (rand(dims,1)-0.5)./10^100;
    s=sprintf('Latent_theta_initialization%d.%d.mat',file_now,iteration); save(s,'theta1');
else
    s=sprintf('Latent_theta_initialization%d.%d.mat',file_now,iteration); load(s);   
end

s=sprintf('Latent%d.%d.mat',file_now,iteration); 
if(exist(s))
    load(s);
end

%BFGS
%--------------------------------------------------------------------------
if(flag_BFGS == 1)
    disp('BFGS');
    flag_alg_on = 1;
    RESULT_BFGS = [];
    latenttab = [];
    cpustart = cputime;
    options = struct('LargeScale','off','HessUpdate','bfgs','GradObj','on','MaxIter',maxiter,'MaxFunEvals',maxiter,'TolX',TolX_val,'TolFun',TolFun_val);
    [X,FVAL,EXITFLAG,OUTPUT]=fminlbfgs(@latent,theta1,options);
    STOPPER = FVAL+1e-4; % log(0.999*exp(FVAL)+0.001*exp(logZ(0*X)));
    s=sprintf('Latent_STOPPER%d.%d.mat',file_now,iteration); save(s,'STOPPER');
    latenttab(:,2) = latenttab(:,2) - cpustart;
    RESULT_BFGS = latenttab;
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_BFGS,1)
      if(RESULT_BFGS(i,1) > RESULT_BFGS(i-1,1))
          RESULT_BFGS(i,1) = RESULT_BFGS(i-1,1);
      end
    end
    %-------------------------------------------------------------------- 
end
%--------------------------------------------------------------------------

%Steepest descent
%--------------------------------------------------------------------------
if(flag_SD == 1)
    disp('Steepest descent');
    flag_alg_on = 2;
    RESULT_SD = [];
    latenttab = [];
    cpustart = cputime;
    options = struct('LargeScale','off','HessUpdate','steepdesc','GradObj','on','MaxIter',maxiter,'MaxFunEvals',maxiter,'TolX',TolX_val,'TolFun',TolFun_val);%,'Display','iter');
    [X,FVAL,EXITFLAG,OUTPUT]=fminlbfgs(@latent,theta1,options);
    latenttab(:,2) = latenttab(:,2) - cpustart;
    RESULT_SD = latenttab;
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_SD,1)
        if(RESULT_SD(i,1) > RESULT_SD(i-1,1))
            RESULT_SD(i,1) = RESULT_SD(i-1,1);
        end
    end
    %--------------------------------------------------------------------
end
%--------------------------------------------------------------------------

%Conjugate gradient
%--------------------------------------------------------------------------
if(flag_CG == 1)
    disp('Conjugate gradient');
    flag_alg_on = 3;
    RESULT_CG = [];
    latenttab = [];
    cpustart = cputime;
    [x fx c] = minimize(theta1, 'latent', -maxiter); %max of maxiter linesearches, if I write - maxiter I have maximum of maxiter function evauations
    latenttab(:,2) = latenttab(:,2) - cpustart;
    RESULT_CG = latenttab;
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_CG,1)
        if(RESULT_CG(i,1) > RESULT_CG(i-1,1))
            RESULT_CG(i,1) = RESULT_CG(i-1,1);
        end
    end
    %--------------------------------------------------------------------
end
%--------------------------------------------------------------------------

%Newton with Hessian inversion 
%--------------------------------------------------------------------------
if(flag_NEWTON_v1 == 1)
    disp('Newton with Hessian inversion v.1');
    RESULT_NEWTON_v1 = [];
    cpustart = cputime;
    theta = theta1;
    [zi,ui] = latent(theta);
    RESULT_NEWTON_v1 = [RESULT_NEWTON_v1;zi,cputime-cpustart,theta'];
    iter = 0;
    flag_iter = 1;
    flag_TolX = 1;
    flag_TolFun = 1;
    while(flag_iter && flag_TolX && flag_TolFun)
        S = eye(dims,dims)*T*LAMBDA;
        if(LAMBDA == 0)
            S = 1e-200*eye(dims,dims);
        end
        u = zeros(dims,1);
        den = 0;
        num = 0;
        for t=1:T
            ztmy = 1e-200;
            ztm = 1e-200;
            utm = zeros(dims,1);
            utmy = zeros(dims,1);
            St1 = ztmy*eye(dims,dims);
            St2 = ztm*eye(dims,dims);
            for m=1:M
                for y=1:Y
                    fi = F{t,y,m};
                    ai = exp(dot(theta,fi));
                    utmy = utmy + ai*fi;
                    ztmy = ztmy + ai;
                end
            end
            utmy = utmy/ztmy;
            for m=1:M
                for y=1:Y
                    fi = F{t,y,m};
                    ai = exp(dot(theta,fi));
                    comp = (utmy - fi)*(utmy - fi)';
                    ratio = (ai/ztmy);
                    St1 = St1 + ratio*comp;
                end
            end
            for m = 1:M
                fi = F{t,YY(t),m};
                ai = exp(dot(theta,fi));
                utm = utm + ai*fi;
                ztm = ztm + ai;
            end
            utm = utm/ztm;
            for m=1:M
                fi = F{t,YY(t),m};
                ai = exp(dot(theta,fi));
                comp = (utm - fi)*(utm - fi)';
                ratio = (ai/ztm);
                St2 = St2 + ratio*comp;
            end       
            S = S + St1 - St2;
            u = u + utmy - utm;
            u = u + LAMBDA*theta;
            num = num + log(ztm) - 0.5*LAMBDA*dot(theta,theta);
            den = den + log(ztmy);
        end
        zi = den - num;
    %     theta_tmp = theta;
        theta = theta - pinv(S)*u;
        RESULT_NEWTON_v1 = [RESULT_NEWTON_v1;zi,cputime-cpustart,theta'];
        iter = iter + 1;
        disp(iter);
        disp(zi);
        %-----------------------Stopping criteria------------------------------
        if(iter > maxiter)
            disp('Optimization completed because the maximum number of iterations was reached');
            flag_iter = 0;
        end
        if(iter > 3)
            if((norm(RESULT_NEWTON_v1(iter)- RESULT_NEWTON_v1(iter-1)) < TolFun_val) && (norm(RESULT_NEWTON_v1(iter-1)- RESULT_NEWTON_v1(iter-2)) < TolFun_val) && (norm(RESULT_NEWTON_v1(iter-2)- RESULT_NEWTON_v1(iter-3)) < TolFun_val))
                disp('Optimization completed because the change of function value is less than the size of the function tolerance');
                flag_TolFun = 0;
            end
        end
    %     if(norm(theta - theta_tmp) < TolX_val)
    %         disp('Optimization completed because the change of parameter vector is less than the size of the parameter tolerance');
    %         flag_TolX = 0;
    %     end
        %----------------------------------------------------------------------
    end
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_NEWTON_v1,1)
        if(RESULT_NEWTON_v1(i,1) > RESULT_NEWTON_v1(i-1,1))
            RESULT_NEWTON_v1(i,1) = RESULT_NEWTON_v1(i-1,1);
        end
    end
    %--------------------------------------------------------------------
end
%--------------------------------------------------------------------------

%Bound with sigma inversion xi = 1.0 Mex version 2
%--------------------------------------------------------------------------
if(flag_BOUND_xi1_Mex_v2 == 1)
    xi = 1.0;   %xi = 2.0;
    disp('Bound with sigma inversion xi = 1.0 Mex v.2');
    RESULT_BOUND_xi1_Mex_v2 = [];
    likelihood_BOUND_xi1_Mex_v2 = [];
    cpustart = cputime;  
    theta = theta1;
    [zi,ui] = latent(theta);
    RESULT_BOUND_xi1_Mex_v2 = [RESULT_BOUND_xi1_Mex_v2;zi,cputime-cpustart,theta'];
    iter = 0;
    flag_iter = 1;
    flag_TolX = 1;
    flag_TolFun = 1;
    while(flag_iter && flag_TolX && flag_TolFun)
        S = eye(dims,dims)*T*LAMBDA;
        if(LAMBDA == 0)
            S = 1e-200*eye(dims,dims);
        end      
        [St,S,den,num,u] = fast_comp_bound(T,M,Y,dims,FM,theta,S,YY,LAMBDA);
        zi = den - num;
        S = (1/xi)*S;
        theta = theta - pinv(S)*u;
        RESULT_BOUND_xi1_Mex_v2 = [RESULT_BOUND_xi1_Mex_v2;zi,cputime-cpustart,theta'];
        save('RESULT_BOUND_xi1_Mex_v2.mat','RESULT_BOUND_xi1_Mex_v2');
        iter = iter + 1;
        disp(iter);
        disp(zi);
        %-----------------------Stopping criteria------------------------------
        if(iter > maxiter)
            disp('Optimization completed because the maximum number of iterations was reached');
            flag_iter = 0;
        end
        if(iter > 3)
            if((norm(RESULT_BOUND_xi1_Mex_v2(iter)- RESULT_BOUND_xi1_Mex_v2(iter-1)) < TolFun_val) && (norm(RESULT_BOUND_xi1_Mex_v2(iter-1)- RESULT_BOUND_xi1_Mex_v2(iter-2)) < TolFun_val) && (norm(RESULT_BOUND_xi1_Mex_v2(iter-2)- RESULT_BOUND_xi1_Mex_v2(iter-3)) < TolFun_val))
                disp('Optimization completed because the change of function value is less than the size of the function tolerance');
                flag_TolFun = 0;
            end
        end
    %     if(norm(theta - theta_tmp) < TolX_val)
    %         disp('Optimization completed because the change of parameter vector is less than the size of the parameter tolerance');
    %         flag_TolX = 0;
    %     end
        %----------------------------------------------------------------------
    end
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_BOUND_xi1_Mex_v2,1)
        if(RESULT_BOUND_xi1_Mex_v2(i,1) > RESULT_BOUND_xi1_Mex_v2(i-1,1))
            RESULT_BOUND_xi1_Mex_v2(i,1) = RESULT_BOUND_xi1_Mex_v2(i-1,1);
        end
    end
    %--------------------------------------------------------------------
end
%--------------------------------------------------------------------------
