% Code by Anna Choromanska and Tony Jebara
% Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

clear all
close all
clc

% GLOBAL VARIABLES
global flag_theta_gen;
global flag_BFGS;
global flag_SD;
global flag_CG;
global flag_NEWTON_v1;
global flag_BOUND_xi1_Mex_v2;
global M;       %number of latent variables
    
% FLAGS
flag_BFGS = 1;
flag_SD = 1;
flag_CG = 1;
flag_NEWTON_v1 = 1;
flag_BOUND_xi1_Mex_v2 = 1;
flag_theta_gen = 0;

% VARIABLES
start_fi = 1;
stop_fi = 1;
start_iter = 1;
stop_iter = 1;
nbofiter = stop_iter - start_iter + 1;
nbofmethods = flag_BFGS + flag_SD + flag_CG + flag_NEWTON_v1 + flag_BOUND_xi1_Mex_v2;
M = 2;  
maxiter = 100000;
TolX_val = 1e-6;
TolFun_val = 1e-6;
p = 0.1;

% %Division to training and testing sets
% for file = start_fi:1:stop_fi
%     file_now = file;
%     disp(file_now); 
%     load('hepatitis.mat');
%     index = [];
%     index = randperm(size(xxx,1));
%     Training_set_x = [];
%     Training_set_y = [];
%     Testing_set_x = [];
%     Testing_set_y = [];
%     y_all = yyy;
%     for i = 1:1:floor(0.9*size(xxx,1))
%         Training_set_x(i,:)= xxx(index(i),:);
%         Training_set_y(i) = yyy(index(i));
%     end
%     for i = 1:1:(size(xxx,1) - floor(0.9*size(xxx,1))-1)
%         Testing_set_x(i,:) = xxx(index(ceil(0.9*size(xxx,1))+i),:);
%         Testing_set_y(i) = yyy(index(ceil(0.9*size(xxx,1))+i));
%     end
%     s=sprintf('Latent_train_test%d.mat',file_now); save(s,'Training_set_x','Training_set_y','Testing_set_x','Testing_set_y','y_all');
% end 

%Latent likelihood training
for file=start_fi:1:stop_fi
    file_now = file;
    disp(file_now);
    s=sprintf('Latent_train_test%d.mat',file_now);  load(s);
    
    for iteration = start_iter:1:stop_iter
        disp(iteration);
        lamb = 0;
        [RESULT_BFGS,RESULT_SD,RESULT_CG,RESULT_NEWTON_v1,RESULT_BOUND_xi1_Mex_v2] = latentlikelihood(Training_set_x,Training_set_y,y_all,lamb,M,maxiter,TolX_val,TolFun_val, file_now, iteration);
        s=sprintf('Latent%d.%d.mat',file_now,iteration); save(s,'RESULT_BFGS','RESULT_SD','RESULT_CG','RESULT_NEWTON_v1','RESULT_BOUND_xi1_Mex_v2');
    end
end

%Latent likelihood testing
for file=start_fi:1:stop_fi
    file_now = file;
    for iteration = start_iter:1:stop_iter
        s=sprintf('Latent_train_test%d.mat',file_now); load(s);
        s=sprintf('Latent%d.%d.mat',file_now,iteration); load(s);
       
        disp(iteration);
        [RESULT_CELL] = data_processing(maxiter,file_now,iteration);
        for i = 1:1:nbofmethods
            likelihood{i} = [];
            for j = 1:1:size(RESULT_CELL{i},1)
                theta_vec = RESULT_CELL{i}(j,3:size(RESULT_CELL{i},2));
                theta_vec = theta_vec';

                [z] = testing_likelihood(theta_vec,Testing_set_x,Testing_set_y,y_all);
                likelihood{i} = [likelihood{i};z];
            end
        end
        
        %smoothing testing likelihood curves
        for i = 1:1:nbofmethods
            for j=2:1:size(likelihood{i},1)
               if(likelihood{i}(j) > likelihood{i}(j-1))
                    likelihood{i}(j) = likelihood{i}(j-1);
               end
            end
        end
        s=sprintf('Latent_test%d.%d.mat',file_now,iteration); save(s,'likelihood');
    end
end

%RESULTS ANALYSIS
for file=start_fi:1:stop_fi
    file_now = file;
    time(:,file_now) = zeros(nbofmethods,1); 
    value(:,file_now) = zeros(nbofmethods,1);
    iterations(:,file_now) = zeros(nbofmethods,1);
    value_t(:,file_now) = zeros(nbofmethods,1);
    percent(:,file_now) = zeros(nbofmethods,1);
    percent_t(:,file_now) = zeros(nbofmethods,1);
    value_conv(:,file_now) = zeros(nbofmethods,1);
    value_conv_var(:,file_now) = zeros(nbofmethods,1);
    time_p(:,file_now) = zeros(nbofmethods,1);
    iterations_p(:,file_now) = zeros(nbofmethods,1);
    value_p(:,file_now) = zeros(nbofmethods,1);
    value_t_p(:,file_now) = zeros(nbofmethods,1);
    for  i = 1:1:nbofmethods
        tab_tmp{i,file_now} = [];
    end
    for iteration = 1:1:nbofiter
        FINAL_TAB = [];
        [FINAL_TAB] = analysis(maxiter,p,file_now,iteration);
        for  i = 1:1:nbofmethods
            value(i,file_now) = value(i,file_now) + FINAL_TAB(i,1);
            time(i,file_now) = time(i,file_now) + FINAL_TAB(i,2);
            iterations(i,file_now) = iterations(i,file_now) + FINAL_TAB(i,3);
            value_t(i,file_now) = value_t(i,file_now) + FINAL_TAB(i,4);
            percent(i,file_now) = percent(i,file_now) + FINAL_TAB(i,5);
            percent_t(i,file_now) = percent_t(i,file_now) + FINAL_TAB(i,6);
            value_conv(i,file_now) = value_conv(i,file_now) + FINAL_TAB(i,7);
            tab_tmp{i,file_now} = [tab_tmp{i,file_now};FINAL_TAB(i,7)];
            time_p(i,file_now) = time_p(i,file_now) + FINAL_TAB(i,8);
            iterations_p(i,file_now) = iterations_p(i,file_now) + FINAL_TAB(i,9);
            value_p(i,file_now) = value_p(i,file_now) + FINAL_TAB(i,10);
            value_t_p(i,file_now) = value_t_p(i,file_now) + FINAL_TAB(i,11);
        end
    end
    time(:,file_now) = time(:,file_now)/nbofiter;
    value(:,file_now) = value(:,file_now)/nbofiter;
    iterations(:,file_now) = round(iterations(:,file_now)/nbofiter);
    value_t(:,file_now) = value_t(:,file_now)/nbofiter;
    percent(:,file_now) = percent(:,file_now)/nbofiter;
    percent_t(:,file_now) = percent_t(:,file_now)/nbofiter;
    value_conv(:,file_now) = value_conv(:,file_now)/nbofiter;
    for  i = 1:1:nbofmethods
        value_conv_var(i,file_now) = var(tab_tmp{i,file_now});
    end
    time_p(:,file_now) = time_p(:,file_now)/nbofiter;
    iterations_p(:,file_now) = iterations_p(:,file_now)/nbofiter;
    value_p(:,file_now) = value_p(:,file_now)/nbofiter;
    value_t_p(:,file_now) = value_t_p(:,file_now)/nbofiter;
end

%RESULT ANALYSIS testing
for file=start_fi:1:stop_fi
    file_now = file;
    value_time_tab{file_now} = zeros(nbofmethods,1); 
    value_iterations_tab{file_now} = zeros(nbofmethods,1); 
    value_convergence_tab{file_now} = zeros(nbofmethods,1); 
    for iteration = 1:1:nbofiter
        s=sprintf('Latent_test%d.%d.mat',file_now,iteration); load(s);
        [value_time,value_iterations, value_convergence] = value_function(maxiter, likelihood, file_now, iteration);
        value_time_tab{file_now} = value_time_tab{file_now} + value_time';
        value_iterations_tab{file_now} = value_iterations_tab{file_now} + value_iterations';
        value_convergence_tab{file_now} = value_convergence_tab{file_now} + value_convergence';
    end
    value_time_tab{file_now} = value_time_tab{file_now}/nbofiter;
    value_iterations_tab{file_now} = value_iterations_tab{file_now}/nbofiter;
    value_convergence_tab{file_now} = value_convergence_tab{file_now}/nbofiter;
end

if(flag_BFGS == 1) fprintf('BFGS');fprintf(1,' \n'); end
if(flag_SD == 1) fprintf('SD');fprintf(1,' \n'); end
if(flag_CG == 1) fprintf('CG');fprintf(1,' \n'); end
if(flag_NEWTON_v1 == 1) fprintf('NEWTON v.1');fprintf(1,' \n'); end
if(flag_BOUND_xi1_Mex_v2 == 1) fprintf('BOUND xi=1.0 Mex v.2');fprintf(1,' \n'); end
fprintf(1,' \n');
fprintf(1,' \n');

%TRAINING RESULTS
fprintf(1,' \n');
disp('time');
for  i = 1:1:nbofmethods
    for file=start_fi:1:stop_fi
        file_now = file;
        fprintf(1,'%2.2f & ', time(i,file_now));
    end
    fprintf(1,' \n');
end
fprintf(1,' \n');

disp('iterations');
for  i = 1:1:nbofmethods
    for file=start_fi:1:stop_fi
        file_now = file;
        fprintf(1,'%2.0f & ', iterations(i,file_now));
    end
    fprintf(1,' \n');
end
fprintf(1,' \n');

fprintf(1,' \n');
disp('value');
for  i = 1:1:nbofmethods
    for file=start_fi:1:stop_fi
        file_now = file;
        fprintf(1,'%2.1f & ', (-1)*value(i,file_now));
    end
    fprintf(1,' \n');
end
fprintf(1,' \n');

%TESTING RESULTS
fprintf(1,' \n');
disp('Testing likelihood in convergence');
for  i = 1:1:nbofmethods
    for file=start_fi:1:stop_fi
        file_now = file;
        iteration = 1;
        s=sprintf('Latent_test%d.%d.mat',file_now,iteration); load(s);
        fprintf(1,'%2.2f & ', (-1)*likelihood{i}(size(likelihood{i},1)));
    end
    fprintf(1,' \n');
end
fprintf(1,' \n');

fprintf(1,' \n');
disp('Testing likelihood in iterations');
for  i = 1:1:nbofmethods
    for file=start_fi:1:stop_fi
        file_now = file;
        iteration = 1;
        s=sprintf('Latent_test%d.%d.mat',file_now,iteration); load(s);
        fprintf(1,'%2.2f & ', (-1)*likelihood{i}(iterations(i,file_now)));
    end
    fprintf(1,' \n');
end
fprintf(1,' \n');

if(nbofmethods == 5)
    linewidth = 2;
    for file=start_fi:1:stop_fi
        for iteration = 1:1:nbofiter
            linewidth = 4;
            file_now = file;
            figure(file+10);
            s=sprintf('Latent%d.%d.mat',file_now,iteration); load(s);
            s=sprintf('Latent_test%d.%d.mat',file_now,iteration); load(s); 
            s=sprintf('Latent_theta_initialization%d.%d.mat',file_now,iteration); load(s); 
            clf;
            plot(log(RESULT_BOUND_xi1_Mex_v2(1:min(size(RESULT_BOUND_xi1_Mex_v2,1),size(likelihood{5},1)),2)),-likelihood{5},'r','LineWidth', linewidth,'Marker','d');hold on;
            ax = axis;
            plot(log(RESULT_NEWTON_v1(1:min(size(RESULT_NEWTON_v1,1),size(likelihood{4},1)),2)),-likelihood{4},'g','LineWidth', linewidth,'Marker','^'); hold on; 
            plot(log(RESULT_BFGS(1:min(size(RESULT_BFGS,1),size(likelihood{1},1)),2)),-likelihood{1},'b','LineWidth', linewidth,'Marker','v'); hold on;
            plot(log(RESULT_CG(1:min(size(RESULT_CG,1),size(likelihood{3},1)),2)),-likelihood{3},'k','LineWidth', linewidth,'Marker','s'); hold on;
            plot(log(RESULT_SD(1:min(size(RESULT_SD,1),size(likelihood{2},1)),2)),-likelihood{2},'color',[0.8,0.8,0.1],'LineWidth', linewidth,'Marker','+'); hold on;
            h=gca;
            %axis([ax(1) ax(2) ax(3) ax(4)])

            set(h,'FontSize',20,'FontWeight','bold');
            title('hepatitis');
            xlabel('log(Time) [sec]');
            ylabel('-log(J(\theta))');
            legend('Bound','Newton','BFGS','Conjugate gradient','Steepest descent','Location','SouthEast');
            set(h,'FontSize',20);
            ss=sprintf('Latentplots%d.%d.eps',file_now,iteration);
            print('-depsc',ss);
        end
    end
end
