// Code by Anna Choromanska and Tony Jebara
// Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "quickmatrix.hpp"
#include "mex.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    //Inputs
    unsigned int t, DD, d_small, n, k, h;
    double *XXM; 
    double *YYM; 
    unsigned int *m; 
    double *theta;
    double *V;
    double *S;
    double *D_vec;
    double *V_in;
    double *S_in;
    double *D_vec_in;
    double LAMBDA;
    double xi;
    
    //Outputs
    double *pzi;
    double *THETA_OUT;
    double *mim;
    
    //Inputs
    //Allocate memory and assign input pointer
    t = (unsigned int)(mxGetScalar(prhs[0]));
    DD = (unsigned int)(mxGetScalar(prhs[1]));
    d_small = (unsigned int)(mxGetScalar(prhs[2]));
    m = (unsigned int *)(mxGetPr(prhs[3]));
    h = (unsigned int)(mxGetScalar(prhs[4]));
    k = (unsigned int)(mxGetScalar(prhs[5]));
    XXM = (double *)(mxGetPr(prhs[6]));
    YYM = (double *)(mxGetPr(prhs[7]));
    theta = (double *)(mxGetPr(prhs[8]));
    V = (double *)(mxGetPr(prhs[9]));
    S = (double *)(mxGetPr(prhs[10]));
    D_vec = (double *)(mxGetPr(prhs[11]));
    V_in = (double *)(mxGetPr(prhs[12]));
    S_in = (double *)(mxGetPr(prhs[13]));
    D_vec_in = (double *)(mxGetPr(prhs[14]));
    LAMBDA = (double)(mxGetScalar(prhs[15]));
    xi = (double)(mxGetScalar(prhs[16]));
    
    //Outputs
    //Allocate memory and assign output pointer
    plhs[0] = mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[1] = mxCreateDoubleMatrix(DD, 1, mxREAL);
    plhs[2] = mxCreateDoubleMatrix(DD, 1, mxREAL);
    pzi = mxGetPr(plhs[0]);
    THETA_OUT = mxGetPr(plhs[1]);
    mim = mxGetPr(plhs[2]);
    
    //Auxilliary variables
	unsigned int i, j, ii, jj, tt, c, u, v, iii;
    Matrix V_mat = MatrixCreate(DD,k);
    Matrix Vt_mat = MatrixCreate(k,DD);
    Matrix S_mat = MatrixCreate(k,k); 
    Vector D_vec_vec = VectorCreate(DD); 
    unsigned int max_m;
    double zti;
    Vector GG = VectorCreate(DD);
    Vector uti = VectorCreate(DD);
    unsigned int addr_x1;
    double *x1;
    Vector logz = VectorCreate(h);
    Vector logzz = VectorCreate(h);
    unsigned int addr;
    double *mi;
    Vector f_vec = VectorCreate(DD);
    unsigned int start, stop;
    unsigned int addr_y1, addr_y2;
    double loga;
    Vector qi = VectorCreate(DD);
    double vt, vv;
    Vector Nq = VectorCreate(DD);
    Vector tmp_dd = VectorCreate(DD);
    Vector tmp_tab1 = VectorCreate(k);
    Vector tmp_tab2 = VectorCreate(k);
    Matrix AA = MatrixCreate(k,k);
    Vector ss = VectorCreate(k);
    double norma_kw;
    double cc;
    double *mimi;
    Vector I = VectorCreate(DD);
    double logzm;
    double tmp;
    double zi;
    double rat1, rat2;
    double suma;
    
    //Auxilliary variables 
    //Memory allocate
    x1 = (double*)mxCalloc(d_small+1, sizeof(double));  
    mi = (double*)mxCalloc(h*DD, sizeof(double));
    mimi = (double*)mxCalloc(h*DD, sizeof(double));
    
    VectorSet(D_vec_vec,0.0);
    VectorSet(Nq,0.0);
    VectorSet(tmp_dd,0.0);
    VectorSet(tmp_tab1,0.0);
    VectorSet(tmp_tab2,0.0);
    VectorSet(ss,0.0);
    VectorSet(I,0.0);
    
    pzi[0] = 0;
    zti = 0;
    
    VectorSet(uti,0.0);
    VectorSet(GG,0.0);
    VectorSet(logz,0.0);
    VectorSet(logzz,0.0);
    
    for (ii = 0; ii <= k-1; ii++)
    {
       for (jj = 0; jj <= DD-1; jj++)
       { 
          V[jj + ii*DD] = V_in[jj + ii*DD]; 
          V_mat->d2[jj][ii] = V[jj + ii*DD];
          Vt_mat->d2[ii][jj] = V[jj + ii*DD];
       }
    }
    
    for (ii = 0; ii <= k-1; ii++)
    {
       for (jj = 0; jj <= k-1; jj++)
       { 
          S[jj + ii*k] = S_in[jj + ii*k]; 
          S_mat->d2[ii][jj] = S[jj + ii*k];
       }
    }
    
    for (jj = 0; jj <= DD-1; jj++)
    { 
        D_vec[jj] = D_vec_in[jj]; 
        D_vec_vec->d[jj] = D_vec[jj];
    }
    
    max_m = m[0];
    for(i = 1; i <= t-1; i++)
    {
        if(max_m < m[i])
        {
            max_m = m[i];
        }
    }              
    
    for(tt = 0; tt <= t-1; tt++)    
    {
        mexPrintf("tt: %d\n", tt); 
        for(c = 0; c <= m[tt]-2; c++)
        {    
            addr_x1 = d_small*(c+tt*max_m);
            
            for(jj = 0; jj <= d_small-1; jj++)
            {
                x1[jj] = XXM[addr_x1];
                addr_x1++;
            }
            x1[d_small] = 1;
        
            if(c == 0)
            {
                for(u = 0; u <= h-1; u++)
                {
                    //mexPrintf("tt: %d\n", tt);                
                    //mexCallMATLAB(0,NULL,0,NULL,"pause");
                    logz->d[u] = log(1e-200);
                    
                    addr = u * DD;
                    for(i = 0; i <= DD-1; i++)
                    {    
                        mi[addr++] = 0;
                    }
                    
                    for(v = 0; v <= h-1; v++)
                    {
                        VectorSet(f_vec,0.0);
                        
                        start = (d_small+1)*(v*h+u);
                        stop = (d_small+1)*(v*h+u+1)-1;
                        
                        iii = 0;
                        for(ii = start; ii <= stop; ii++)
                        {                
                            f_vec->d[ii] = x1[iii]; 
                            iii++;
                        }
                        
                        addr_y1 = c+tt*max_m;
                        addr_y2 = c+1+tt*max_m;
                     
                        if((v == YYM[addr_y1]-1) && (u == YYM[addr_y2]-1))
                        {
                            VectorAdd(GG,f_vec);
                        }
                        
                        loga = 0;
                        for(ii = 0; ii <= DD-1; ii++)
                        {                
                            loga += theta[ii] * f_vec->d[ii];
                        }

                        VectorSet(qi,0.0);
                        addr = u * DD;
                        for(ii = 0; ii <= DD-1; ii++)
                        {                
                            qi->d[ii] = mi[addr];
                            addr++;
                        }
                        VectorSubtract(qi,f_vec);
                        if (loga==logz->d[u])
                        {
                            VectorScale(qi,0.5);
                        }
                        else
                        {
                            vt = loga - logz->d[u];
                            vv = tanh(0.5*vt);
                            vv /= (2*vt);
                            vv = sqrt(vv);
                            VectorScale(qi,vv);
                        }     
                        
                        VectorMove(Nq,qi);

                        MatrixVectorMultiply(tmp_tab1,Vt_mat,Nq);
                        MatrixVectorMultiply(tmp_dd,V_mat,tmp_tab1);
                        VectorSubtract(Nq,tmp_dd);

                        for (ii=0; ii<k; ii++)
                        {
                            for (jj=0; jj<k; jj++)
                            {
                                AA->d2[ii][jj] = S_mat->d2[ii][jj] + tmp_tab1->d[ii]*tmp_tab1->d[jj];
                            }
                        }

                        MatrixEigenvectorsSymTony(AA,ss->d);

                        VectorSet((Vector) S_mat,0.0);
                        for (ii=0; ii<k; ii++) 
                        {
                            S_mat->d2[ii][ii] = ss->d[ii];
                        }

                        MatrixMultiply(Vt_mat,AA,Vt_mat);
                        MatrixTranspose(AA);
                        MatrixMultiply(V_mat,V_mat,AA);

                        norma_kw = VectorDot(Nq,Nq);
                        
                        double min_v = norma_kw;
                        unsigned int min_i = k;
                        for (ii=0; ii<k; ii++) 
                        {
                            if (S_mat->d2[ii][ii] < min_v)
                            {
                                min_v = S_mat->d2[ii][ii];
                                min_i = ii;
                            }
                        }

                        if (min_i<k) 
                        {
                            for (jj=0; jj<DD; jj++) 
                            {
                                tmp_dd->d[jj] = fabs(V_mat->d2[jj][min_i]);
                            }
                            suma = VectorSum(tmp_dd);
                            cc = suma*S_mat->d2[min_i][min_i];
                            VectorScale(tmp_dd,cc);
                            VectorAdd(D_vec_vec,tmp_dd); 
                            S_mat->d2[min_i][min_i] = norma_kw;
                            norma_kw = 1.0/sqrt(norma_kw);
                            VectorScale(Nq,norma_kw);
                            for (jj=0; jj<(DD); jj++)
                            {
                                V_mat->d2[jj][min_i]  = Nq->d[jj];
                                Vt_mat->d2[min_i][jj] = Nq->d[jj];
                            } 
                        }
                        else
                        {
                            for (jj=0; jj<DD; jj++) 
                            {
                                tmp_dd->d[jj] = fabs(Nq->d[jj]);
                            }
                            cc = VectorSum(tmp_dd);
                            VectorScale(tmp_dd,cc);
                            VectorAdd(D_vec_vec,tmp_dd);
                        }
                        
                        if(logz->d[u] > loga)
                        {
                            rat1 = 1/(1 + exp(loga-logz->d[u]));
                            rat2 = 1 - rat1;
                        }
                        else
                        {
                            rat2 = 1/(exp(logz->d[u]-loga) + 1);
                            rat1 = 1 - rat2;   
                        }                   

                        addr = u * DD;
                        for (ii = 0; ii <= DD-1; ii++)
                        {
                            mi[addr] = rat1*mi[addr] + rat2*f_vec->d[ii];
                            addr++;
                        }

                        if(loga < logz->d[u])
                        {   
                            logz->d[u] = logz->d[u] + log(1 + exp(loga - logz->d[u]));
                        }
                        else
                        {
                            logz->d[u] = loga + log(1 + exp(logz->d[u] - loga));  
                        }          
                    }                    
                }  
            }
            else if(c < m[tt]-2)
            {                        
                addr = 0;
                for(ii = 0; ii <= h-1; ii++)
                {
                    logzz->d[ii] = logz->d[ii];
                
                    for(jj = 0; jj <= DD-1; jj++)
                    {
                        mimi[addr] = mi[addr];
                        addr++;
                    }
                }
                
                for(u = 0; u <= h-1; u++)
                {
                    logz->d[u] = log(1e-200);
                    
                    addr = u * DD;
                    for(i = 0; i <= DD-1; i++)
                    {    
                        mi[addr] = 0;
                        addr++;
                    }
                
                    for(v = 0; v <= h-1; v++)
                    {        
                        VectorSet(f_vec,0.0);
                        
                        start = (d_small+1)*(v*h+u);
                        stop = (d_small+1)*(v*h+u+1)-1;
                        
                        iii = 0;
                        for(ii = start; ii <= stop; ii++)
                        {                
                            f_vec->d[ii] = x1[iii];
                            iii++;
                        }
                        
                        addr_y1 = c+tt*max_m;
                        addr_y2 = c+1+tt*max_m;
                        if((v == YYM[addr_y1]-1) && (u == YYM[addr_y2]-1))
                        {
                            VectorAdd(GG,f_vec);
                        }
                        
                        addr = v * DD;
                        for(ii = 0; ii <= DD-1; ii++)
                        {    
                            I->d[ii] = f_vec->d[ii] + mimi[addr];
                            addr++;
                        }
                                           
                        loga = 0;
                        for(ii = 0; ii <= DD-1; ii++)
                        {                
                            loga += theta[ii] * f_vec->d[ii];
                        }
                        loga += logzz->d[v];
                        
                        VectorSet(qi,0.0);
                        addr = u * DD;
                        for(ii = 0; ii <= DD-1; ii++)
                        {                
                            qi->d[ii] = mi[addr];
                            addr++;
                        }
                        VectorSubtract(qi,I);
                        if (loga==logz->d[u])
                        {
                            VectorScale(qi,0.5);
                        }
                        else
                        {
                            vt = loga - logz->d[u];
                            vv = tanh(0.5*vt);
                            vv /= (2*vt);
                            vv = sqrt(vv);
                            VectorScale(qi,vv);
                        }     
                        
                        VectorMove(Nq,qi);
                        
                        MatrixVectorMultiply(tmp_tab1,Vt_mat,Nq);
                        MatrixVectorMultiply(tmp_dd,V_mat,tmp_tab1);
                        VectorSubtract(Nq,tmp_dd);

                        for (ii=0; ii<k; ii++)
                        {
                            for (jj=0; jj<k; jj++)
                            {
                                AA->d2[ii][jj] = S_mat->d2[ii][jj] + tmp_tab1->d[ii]*tmp_tab1->d[jj];
                            }
                        }

                        MatrixEigenvectorsSymTony(AA,ss->d);

                        VectorSet((Vector) S_mat,0.0);
                        for (ii=0; ii<k; ii++) 
                        {
                            S_mat->d2[ii][ii] = ss->d[ii];
                        }

                        MatrixMultiply(Vt_mat,AA,Vt_mat);
                        MatrixTranspose(AA);
                        MatrixMultiply(V_mat,V_mat,AA);

                        norma_kw = VectorDot(Nq,Nq);
                        
                        double min_v = norma_kw;
                        unsigned int min_i = k;
                        for (ii=0; ii<k; ii++) 
                        {
                            if (S_mat->d2[ii][ii] < min_v)
                            {
                                min_v = S_mat->d2[ii][ii];
                                min_i = ii;
                            }
                        }

                        if (min_i<k) 
                        {
                            for (jj=0; jj<DD; jj++) 
                            {
                                tmp_dd->d[jj] = fabs(V_mat->d2[jj][min_i]);
                            }
                            suma = VectorSum(tmp_dd);
                            cc = suma*S_mat->d2[min_i][min_i];
                            VectorScale(tmp_dd,cc);
                            VectorAdd(D_vec_vec,tmp_dd); 
                            S_mat->d2[min_i][min_i] = norma_kw;
                            norma_kw = 1.0/sqrt(norma_kw);
                            VectorScale(Nq,norma_kw);
                            for (jj=0; jj<(DD); jj++)
                            {
                                V_mat->d2[jj][min_i]  = Nq->d[jj];
                                Vt_mat->d2[min_i][jj] = Nq->d[jj];
                            } 
                        }
                        else
                        {
                            for (jj=0; jj<DD; jj++) 
                            {
                                tmp_dd->d[jj] = fabs(Nq->d[jj]);
                            }
                            cc = VectorSum(tmp_dd);
                            VectorScale(tmp_dd,cc);
                            VectorAdd(D_vec_vec,tmp_dd);
                        }
                        
                        if(logz->d[u] > loga)
                        {
                            rat1 = 1/(1 + exp(loga-logz->d[u]));
                            rat2 = 1 - rat1;
                        }
                        else
                        {
                            rat2 = 1/(exp(logz->d[u]-loga) + 1);
                            rat1 = 1 - rat2;   
                        }                   

                        addr = u * DD;
                        for (ii = 0; ii <= DD-1; ii++)
                        {
                            mi[addr] = rat1*mi[addr] + rat2*I->d[ii];
                            addr++;
                        }

                        if(loga < logz->d[u])
                        {   
                            logz->d[u] = logz->d[u] + log(1 + exp(loga - logz->d[u]));
                        }
                        else
                        {
                            logz->d[u] = loga + log(1 + exp(logz->d[u] - loga));  
                        }                           
                    }                    
                }
            }
            else
            {
                logzm = log(1e-200);
                
                for(jj = 0; jj <= DD-1; jj++)
                {
                    mim[jj] = 0;
                }                
                
                for(u = 0; u <= h-1; u++)
                {
                    for(v = 0; v <= h-1; v++)
                    {   
                        VectorSet(f_vec,0.0);
                        
                        start = (d_small+1)*(v*h+u);
                        stop = (d_small+1)*(v*h+u+1)-1;
                        
                        iii = 0;
                        for(ii = start; ii <= stop; ii++)
                        {                
                            f_vec->d[ii] = x1[iii];
                            iii++;
                        }
                        
                        addr_y1 = c+tt*max_m;
                        addr_y2 = c+1+tt*max_m;
                        if((v == YYM[addr_y1]-1) && (u == YYM[addr_y2]-1))
                        {
                            VectorAdd(GG,f_vec);
                        }
                        
                        addr = v * DD;
                        for(ii = 0; ii <= DD-1; ii++)
                        {    
                            I->d[ii] = f_vec->d[ii] + mi[addr++];
                        }
                        
                        loga = 0;
                        for(ii = 0; ii <= DD-1; ii++)
                        {                
                            loga += theta[ii] * f_vec->d[ii];
                        }
                        loga += logz->d[v];
                        
                        VectorSet(qi,0.0);
                        for(ii = 0; ii <= DD-1; ii++)
                        {                
                            qi->d[ii] = mim[ii];
                        }
                        VectorSubtract(qi,I);
                        if (loga==logzm)
                        {
                            VectorScale(qi,0.5);
                        }
                        else
                        {
                            vt = loga - logzm;
                            vv = tanh(0.5*vt);
                            vv /= (2*vt);
                            vv = sqrt(vv);
                            VectorScale(qi,vv);
                        }     
                        
                        VectorMove(Nq,qi);
                        
                        MatrixVectorMultiply(tmp_tab1,Vt_mat,Nq);
                        MatrixVectorMultiply(tmp_dd,V_mat,tmp_tab1);
                        VectorSubtract(Nq,tmp_dd);

                        for (ii=0; ii<k; ii++)
                        {
                            for (jj=0; jj<k; jj++)
                            {
                                AA->d2[ii][jj] = S_mat->d2[ii][jj] + tmp_tab1->d[ii]*tmp_tab1->d[jj];
                            }
                        }

                        MatrixEigenvectorsSymTony(AA,ss->d);

                        VectorSet((Vector) S_mat,0.0);
                        for (ii=0; ii<k; ii++) 
                        {
                            S_mat->d2[ii][ii] = ss->d[ii];
                        }

                        MatrixMultiply(Vt_mat,AA,Vt_mat);
                        MatrixTranspose(AA);
                        MatrixMultiply(V_mat,V_mat,AA);

                        norma_kw = VectorDot(Nq,Nq);
                        
                        double min_v = norma_kw;
                        unsigned int min_i = k;
                        for (ii=0; ii<k; ii++) 
                        {
                            if (S_mat->d2[ii][ii] < min_v)
                            {
                                min_v = S_mat->d2[ii][ii];
                                min_i = ii;
                            }
                        }

                        if (min_i<k) 
                        {
                            for (jj=0; jj<DD; jj++) 
                            {
                                tmp_dd->d[jj] = fabs(V_mat->d2[jj][min_i]);
                            }
                            suma = VectorSum(tmp_dd);
                            cc = suma*S_mat->d2[min_i][min_i];
                            VectorScale(tmp_dd,cc);
                            VectorAdd(D_vec_vec,tmp_dd); 
                            S_mat->d2[min_i][min_i] = norma_kw;
                            norma_kw = 1.0/sqrt(norma_kw);
                            VectorScale(Nq,norma_kw);
                            for (jj=0; jj<(DD); jj++)
                            {
                                V_mat->d2[jj][min_i]  = Nq->d[jj];
                                Vt_mat->d2[min_i][jj] = Nq->d[jj];
                            } 
                        }
                        else
                        {
                            for (jj=0; jj<DD; jj++) 
                            {
                                tmp_dd->d[jj] = fabs(Nq->d[jj]);
                            }
                            cc = VectorSum(tmp_dd);
                            VectorScale(tmp_dd,cc);
                            VectorAdd(D_vec_vec,tmp_dd);
                        }
                        
                        if(logzm > loga)
                        {
                            rat1 = 1/(1 + exp(loga-logzm));
                            rat2 = 1 - rat1;
                        }
                        else
                        {
                            rat2 = 1/(exp(logzm-loga) + 1);
                            rat1 = 1 - rat2;   
                        }                   
        
                        for (ii = 0; ii <= DD-1; ii++)
                        {
                            mim[ii] = rat1*mim[ii] + rat2*I->d[ii];
                        }

                        if(loga < logzm)
                        {   
                            logzm = logzm + log(1 + exp(loga - logzm));
                        }
                        else
                        {
                            logzm = loga + log(1 + exp(logzm - loga));  
                        }
                    } 
                }
            }        
        }
        
        for (ii = 0; ii <= DD-1; ii++)
        {            
            mim[ii] += LAMBDA * theta[ii];
        }
        
        tmp = 0;
        for (ii = 0; ii <= DD-1; ii++)
        {            
            tmp += theta[ii] * theta[ii];
        }
        
        zi = logzm + 0.5 * LAMBDA * tmp;
        
        for (ii = 0; ii <= DD-1; ii++)
        {            
            uti->d[ii] += mim[ii];
        }
        
        zti += zi;
    }    
    
    for (ii = 0; ii <= DD-1; ii++)
    {            
        mim[ii] = uti->d[ii] - GG->d[ii];
    }
    
    tmp = 0;
    for (ii = 0; ii <= DD-1; ii++)
    {            
        tmp += theta[ii] * GG->d[ii];
    }
    
    zi = zti - tmp;
    pzi[0] = zi;
    
    //(A + BC)^-1 = A^-1 - A^-1 B (I + C A-1 B)^-1 C A^-1.
    // Invert lowrank&diagonal covariance prior to qi multiply
    //                       ( A + B   C    )
    // Use Woodbury above on ( D + V'*(S*V) )^{-1}
    // where A->D B->V' and C->SV
    // so inv = D^-1 - D^-1 V' (I+ (S*V) D^-1 V' )^-1 S*V  D^-1

    for (jj=0; jj<DD; jj++) 
    {
        tmp_dd->d[jj] = mim[jj]/D_vec_vec->d[jj];
    }
    for (ii=0; ii<k; ii++) 
    {
        for (jj=0; jj<DD; jj++) 
        {
            Vt_mat->d2[ii][jj] /= D_vec_vec->d[jj];
        }
    }
    MatrixMultiply(AA,Vt_mat,V_mat);
    MatrixMultiply(AA,S_mat,AA);
    for (ii=0; ii<k; ii++)
    {
        AA->d2[ii][ii] += 1.0;
    }
    MatrixInvert(AA);
    MatrixMultiply(AA,AA,S_mat);
    MatrixVectorMultiply(tmp_tab1,Vt_mat,tmp_dd);
    MatrixVectorMultiply(tmp_tab2,AA,tmp_tab1);
    for (ii=0; ii<k; ii++) 
    {
        for (jj=0; jj<DD; jj++) 
        {
            Vt_mat->d2[ii][jj] = V_mat->d2[jj][ii];
        }
    }
    MatrixVectorMultiply(Nq,V_mat,tmp_tab2);
    for (jj=0; jj<DD; jj++) 
    {
        Nq->d[jj] = mim[jj]/D_vec_vec->d[jj] - Nq->d[jj]/D_vec_vec->d[jj];
    }
    for (jj=0; jj<DD; jj++) 
    {
        THETA_OUT[jj] = theta[jj] - xi*Nq->d[jj];
    }
    
    //Delete
    VectorFree(GG);
    VectorFree(uti);
    VectorFree(f_vec);
    VectorFree(qi);
    VectorFree(Nq);
    VectorFree(I);
    MatrixFree(V_mat);
    MatrixFree(Vt_mat);
    VectorFree(tmp_dd);
    VectorFree(tmp_tab1);
    VectorFree(tmp_tab2);
    MatrixFree(S_mat);
    MatrixFree(AA);
    VectorFree(ss);
    VectorFree(D_vec_vec);
    VectorFree(logz);
    VectorFree(logzz);
}   
