% Code by Anna Choromanska and Tony Jebara
% Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

function [zi,ui] = logZTree_all(THETA)
global d;
global m;
global t;
global h;
global XX;
global YY;
global LAMBDA;
global logZtab;
global maxiter;
global flag_alg_on;

if(flag_alg_on == 1)
    save('RESULT_LBFGS.mat','logZtab');
elseif(flag_alg_on == 2)
    save('RESULT_SD.mat','logZtab');
elseif(flag_alg_on == 3)
    save('RESULT_CG.mat','logZtab');
end

zti = 0.0;
zti_tmp = 0.0;
uti=zeros(d,1);
d_small = size(XX{t},2);
GG=zeros(1,d);
for tt = 1:1:t
    disp(tt);
    disp(size(logZtab,1));
    %one pass of computing the bound
    for c = 1:1:m(tt)-1
        x1=[XX{tt}(c,:) 1]; 
        if(c == 1)  %computing bound parameters for a leaf
            for u = 1:1:h
                z(u) = 1e-200;
                mi{u} = zeros(d,1);
                for v = 1:1:h
                    f = sparse(d,1);
                    start = (d_small+1)*((v-1)*h+u-1)+1;
                    stop = (d_small+1)*((v-1)*h+u);
                    f(start:stop,1)=x1;
                    if ((v==YY{tt}(c)) && (u==YY{tt}(c+1)))
                        GG = GG + f';
                    end
                    a = exp(THETA'*f);
                    comp1 = (a/(z(u) + a)).*f;
                    comp2 = (z(u)/(z(u) + a)).*mi{u};
                    mi{u} = comp1 + comp2;
                    z(u) = z(u) + a;
                end
            end
        elseif(c~=m(tt)-1)  %computing bound parameters for a non-leaf and non-root
            zz = z;
            mimi = mi;
            for u = 1:1:h
                z(u) = 1e-200;
                mi{u} = zeros(d,1);
                for v = 1:1:h
                    f = sparse(d,1);
                    start = (d_small+1)*((v-1)*h+u-1)+1;
                    stop = (d_small+1)*((v-1)*h+u);
                    f(start:stop,1)=x1;
                    if ((v==YY{tt}(c)) && (u==YY{tt}(c+1)))
                        GG = GG + f';
                    end
                    I = f + mimi{v};
                    a = exp(THETA'*f);
                    a = a*zz(v);
                    comp1 = (a/(z(u) + a)).*I;
                    comp2 = (z(u)/(z(u) + a)).*mi{u};
                    mi{u} = comp1 + comp2;
                    z(u) = z(u) + a;
                end
            end
        else  %computing bound parameters for a root
            zm = 1e-200;
            mim = zeros(d,1);
            for u = 1:1:h
                for v = 1:1:h
                    f = sparse(d,1);
                    start = (d_small+1)*((v-1)*h+u-1)+1;
                    stop = (d_small+1)*((v-1)*h+u);
                    f(start:stop,1)=x1;
                    if ((v==YY{tt}(c)) && (u==YY{tt}(c+1)))
                        GG = GG + f';
                    end
                    I = f + mi{v};
                    a = exp(THETA'*f);
                    a = a*z(v);
                    comp1 = (a/(zm + a)).*I;
                    comp2 = (zm/(zm + a)).*mim;
                    mim = comp1 + comp2;
                    zm = zm + a;
                end
            end
        end
    end
    ui = mim + LAMBDA*THETA;
    zi = log(zm) + 0.5*LAMBDA*dot(THETA,THETA);
    uti = uti+ui;
    zti = zti+zi;
    
    zi_tmp = log(zm);
    zti_tmp = zti_tmp + zi_tmp;
end
ui = uti-GG';
zi = zti-dot(THETA,GG');

zi_tmp = zti_tmp - dot(THETA,GG');

if(size(logZtab,1) > maxiter)
    zi = logZtab(size(logZtab,1),1);
    ui = zeros(d,1);
end

logZtab = [logZtab;zi,cputime,THETA'];

disp(zi);

