% Code by Anna Choromanska and Tony Jebara
% Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

function [RESULT_LBFGS,RESULT_SD,RESULT_CG,RESULT_BOUND_Matlab_v1,RESULT_BOUND_Mex_v2] = realbound(XX,YY,lamb,maxiter,TolX_val,TolFun_val,file_now,ll,iteration)
% GLOBALS
global LAMBDA;
global logZtab;
global d;
global m;
global t;
global h;
global XX;
global YY;
global maxiter;
global flag_alg_on;
global flag_theta_gen;
global flag_LBFGS;
global flag_SD;
global flag_CG;
global flag_BOUND_Matlab_v1;
global flag_BOUND_Mex_v2;

% OUTPUT VARIABLES INITIALIZATION
RESULT_LBFGS = [1,1];
RESULT_SD = [1,1];
RESULT_CG = [1,1];
RESULT_BOUND_Matlab_v1 = [1,1];
RESULT_BOUND_Mex_v2 = [1,1];

% VARIABLES
LAMBDA=lamb;
t = size(XX,2);
y_t = [];
for tt=1:1:t
    tmp = [];
    tmp = zeros(size(YY{tt},1),1);    
    tmp(1:size(YY{tt},1),1) = YY{tt}(:,1);
    y_t = [y_t;tmp];
end
h = length(unique(y_t));
for tt = 1:1:t
    m(tt) = size(XX{tt},1);
end
d = size(XX{t},2);
d_small = size(XX{t},2);
DD =(d+1)*h*h;
d = DD;
disp('Model');
disp('[t m d h]');
disp([t m d h]);
k = ceil(log10(DD));
[XXM] = ArrayToMatrixconversion(XX);
[YYM] = ArrayToMatrixconversion(YY);
theta1 = zeros(d,1);
if(flag_theta_gen == 1)
    theta1 = (rand(DD,1)-0.5)./10^10; 
    s=sprintf('Trees_theta_initialization%d.%d.%d.mat',file_now,ll,iteration); save(s,'theta1'); %load(s);  
else
    s=sprintf('Trees_theta_initialization%d.%d.%d.mat',file_now,ll,iteration); load(s);   
end

s=sprintf('Trees%d.%d.%d.mat',file_now,ll,iteration); 
if(exist(s))
    load(s);
end

%LBFGS
%--------------------------------------------------------------------------
if(flag_LBFGS == 1)
    disp('LBFGS');
    flag_alg_on = 1;
    RESULT_LBFGS = [];
    logZtab = [];
    cpustart = cputime;
    options = struct('HessUpdate','lbfgs','StoreN',k,'GradObj','on','MaxFunEvals',maxiter,'TolX',TolX_val,'TolFun',TolFun_val);
    [X,FVAL,EXITFLAG,OUTPUT]=fminlbfgs(@logZTree_all,theta1,options);
    STOPPER = FVAL+1e-4; % log(0.999*exp(FVAL)+0.001*exp(logZ(0*X)));
    s=sprintf('Trees_STOPPER%d.%d.%d.mat',file_now,ll,iteration); save(s,'STOPPER');
    logZtab(:,2) = logZtab(:,2) - cpustart;
    RESULT_LBFGS = logZtab;
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_LBFGS,1)
        if(RESULT_LBFGS(i,1) > RESULT_LBFGS(i-1,1))
            RESULT_LBFGS(i,1) = RESULT_LBFGS(i-1,1);
        end
    end
    %-------------------------------------------------------------------- 
end
%--------------------------------------------------------------------------

%Steepest descent
%--------------------------------------------------------------------------
if(flag_SD == 1) 
    disp('Steepest descent');
    flag_alg_on = 2;
    RESULT_SD = [];
    logZtab = [];
    cpustart = cputime;
    options = struct('LargeScale','off','HessUpdate','steepdesc','GradObj','on','MaxFunEvals',maxiter,'TolX',TolX_val,'TolFun',TolFun_val);%,'Display','iter');
    [X,FVAL,EXITFLAG,OUTPUT]=fminlbfgs(@logZTree_all,theta1,options);
    logZtab(:,2) = logZtab(:,2) - cpustart;
    RESULT_SD = logZtab;
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_SD,1)
        if(RESULT_SD(i,1) > RESULT_SD(i-1,1))
            RESULT_SD(i,1) = RESULT_SD(i-1,1);
        end
    end
    %--------------------------------------------------------------------    
end
%--------------------------------------------------------------------------

%Conjugate gradient
%--------------------------------------------------------------------------
if(flag_CG == 1)
    disp('Conjugate gradient');
    flag_alg_on = 3;
    RESULT_CG = [];
    logZtab = [];
    cpustart = cputime;
    [x fx c] = minimize(theta1, 'logZTree_all', -maxiter); %max of maxiter linesearches, if I write - maxiter I have maximum of maxiter function evauations
    logZtab(:,2) = logZtab(:,2) - cpustart;
    RESULT_CG = logZtab;
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_CG,1)
        if(RESULT_CG(i,1) > RESULT_CG(i-1,1))
            RESULT_CG(i,1) = RESULT_CG(i-1,1);
        end
    end
    %--------------------------------------------------------------------    
end
%--------------------------------------------------------------------------

%Bound with sigma inversion xi = 1.0 Matlab implementation (equivalent to Mex implementation however slower)
%--------------------------------------------------------------------------
%k = DD;
if(flag_BOUND_Matlab_v1 == 1)
    xi = 1.0;
    disp('Bound with sigma inversion xi = 1.0 Matlab v.1');
    RESULT_BOUND_Matlab_v1 = [];
    cpustart = cputime;
    THETA = theta1;
    iter = 0;
    flag_iter = 1;
    flag_TolX = 1;
    flag_TolFun = 1;
    while(flag_iter && flag_TolX && flag_TolFun)
        zti = 0.0;
        uti=zeros(d,1);
        %------------------------------------------------------------------
        V = zeros(DD,k);
        V_tmp = rand(k,k);
        [Q,R] = qr(V_tmp);
        V(1:k,1:k) = Q; 
        D_vec = ones(DD,1).*(t*LAMBDA);
        S = zeros(k,k);
        %------------------------------------------------------------------
        GG=zeros(1,d);
        %S_dod = eye(DD,DD)*t*LAMBDA;
        for tt = 1:1:t
            %one pass of computing the bound
            for c = 1:1:m(tt)-1
                %St_dod = 1e-200*eye(DD,DD);
                x1=[XX{tt}(c,:) 1]; 
                if(c == 1)  %computing bound parameters for a leaf
                    for u = 1:1:h
                        %z(u) = 1e-200;
                        logz(u) = log(1e-200);
                        mi{u} = zeros(d,1);
                        for v = 1:1:h
                            f = sparse(d,1);
                            start = (d_small+1)*((v-1)*h+u-1)+1;
                            stop = (d_small+1)*((v-1)*h+u);
                            f(start:stop,1)=x1;
                            if ((v==YY{tt}(c)) && (u==YY{tt}(c+1)))
                                GG = GG + f';
                            end
                            %a = exp(dot(THETA,f));
                            loga = dot(THETA,f);
                            %if(a==z(u))
                            if(loga == logz(u))
                                qi = 0.5*(mi{u}-f);
                            else
                                %qi = sqrt( tanh(0.5*(log(a) - log(z(u))))/(2*(log(a) - log(z(u)))) )*(mi{u}-f);
                                qi = sqrt( tanh(0.5*(loga - logz(u)))/(2*(loga - logz(u))) )*(mi{u}-f);
                            end
                            %St_dod = St_dod + qi*qi';
                            %------------------------------------------------------------------
                            Nq = qi;
                            tmp_tab1 = V'*Nq;
                            tmp_dd = V*tmp_tab1; 
                            Nq = Nq - tmp_dd;
                            S = S + tmp_tab1*tmp_tab1';
                            [Vq,Sq] = eig(S);
                            S = Sq;
                            V = V*Vq;               
                            norma_kw = norm(Nq)^2;
                            S_tmp = [diag(S); norma_kw];
                            [value,index] = min(S_tmp);
                            if(index <= k)
                                cc = S(index,index)*sum(abs(V(:,index)));
                                D_vec = D_vec + cc*(abs(V(:,index)));
                                S(index,index) = norma_kw;        
                                Nq = Nq/sqrt(norma_kw);
                                V(:,index) = Nq;
                            else
                                cc = sum(abs(Nq));
                                D_vec = D_vec + cc*abs(Nq);
                            end
                            %------------------------------------------------------------------
%                             comp1 = (a/(z(u) + a)).*f;
%                             comp2 = (z(u)/(z(u) + a)).*mi{u};
%                             mi{u} = comp1 + comp2;
                            if(logz(u)>loga)
                                rat1 = 1.0/(1.0+exp(loga-logz(u)));
                                rat2 = 1.0-rat1;
                            else
                                rat2 = 1.0/(exp(logz(u)-loga)+1.0);
                                rat1 = 1.0-rat2;
                            end   
                            mi{u} = rat1*mi{u} + rat2*f;
                            %z(u) = z(u) + a;
                             if(loga<logz(u)) 
                                logz(u) = logz(u) + log(1.0+exp(loga-logz(u)));
                            else
                                logz(u) = loga + log(1.0+exp(logz(u)-loga));
                            end
                        end
                    end
                elseif(c~=m(tt)-1)  %computing bound parameters for a non-leaf and non-root
                    %zz = z;
                    logzz = logz;
                    mimi = mi;
                    for u = 1:1:h
                        %z(u) = 1e-200;
                        logz(u) = log(1e-200);
                        mi{u} = zeros(d,1);
                        for v = 1:1:h
                            f = sparse(d,1);
                            start = (d_small+1)*((v-1)*h+u-1)+1;
                            stop = (d_small+1)*((v-1)*h+u);
                            f(start:stop,1)=x1;
                            if ((v==YY{tt}(c)) && (u==YY{tt}(c+1)))
                                GG = GG + f';
                            end
                            I = f + mimi{v};
                            %a = exp(dot(THETA,f));
                            loga = dot(THETA,f);
                            %a = a*zz(v);
                            loga = loga + logzz(v);
                            %if(a==z(u))
                            if(loga == logz(u))
                                qi = 0.5*(mi{u}-I);
                            else
                                %qi = sqrt( tanh(0.5*(log(a) - log(z(u))) )/(2*(log(a) - log(z(u)))))*(mi{u}-I);
                                qi = sqrt( tanh(0.5*(loga - logz(u)) )/(2*(loga - logz(u))))*(mi{u}-I);
                            end
                            %St_dod = St_dod + qi*qi';
                            %------------------------------------------------------------------
                            Nq = qi;
                            tmp_tab1 = V'*Nq;
                            tmp_dd = V*tmp_tab1; 
                            Nq = Nq - tmp_dd;
                            S = S + tmp_tab1*tmp_tab1';
                            [Vq,Sq] = eig(S);
                            S = Sq;
                            V = V*Vq;               
                            norma_kw = norm(Nq)^2;
                            S_tmp = [diag(S); norma_kw];
                            [value,index] = min(S_tmp);
                            if(index <= k)
                                cc = S(index,index)*sum(abs(V(:,index)));
                                D_vec = D_vec + cc*(abs(V(:,index)));
                                S(index,index) = norma_kw;        
                                Nq = Nq/sqrt(norma_kw);
                                V(:,index) = Nq;
                            else
                                cc = sum(abs(Nq));
                                D_vec = D_vec + cc*abs(Nq);
                            end
                            %------------------------------------------------------------------
%                             comp1 = (a/(z(u) + a)).*I;
%                             comp2 = (z(u)/(z(u) + a)).*mi{u};
%                             mi{u} = comp1 + comp2;
                           if(logz(u)>loga)
                                rat1 = 1.0/(1.0+exp(loga-logz(u)));
                                rat2 = 1.0-rat1;
                            else
                                rat2 = 1.0/(exp(logz(u)-loga)+1.0);
                                rat1 = 1.0-rat2;
                            end   
                            mi{u} = rat1*mi{u} + rat2*I;
                            %z(u) = z(u) + a;
                            if(loga<logz(u)) 
                                logz(u) = logz(u) + log(1.0+exp(loga-logz(u)));
                            else
                                logz(u) = loga + log(1.0+exp(logz(u)-loga));
                            end
                        end
                    end
                else  %computing bound parameters for a root
                    %zm = 1e-200;
                    logzm = log(1e-200);
                    mim = zeros(d,1);
                    for u = 1:1:h
                        for v = 1:1:h
                            f = sparse(d,1);
                            start = (d_small+1)*((v-1)*h+u-1)+1;
                            stop = (d_small+1)*((v-1)*h+u);
                            f(start:stop,1)=x1;
                            if ((v==YY{tt}(c)) && (u==YY{tt}(c+1)))
                                GG = GG + f';
                            end
                            I = f + mi{v};
                            %a = exp(dot(THETA,f));
                            loga = dot(THETA,f);
                            %a = a*z(v);
                            loga = loga + logz(v);
                            %if(a==zm)
                            if(loga == logzm)
                                qi = 0.5*(mim-I);
                            else
                                %qi = sqrt( tanh(0.5*(log(a) - log(zm)))/(2*(log(a) - log(zm))) )*(mim-I);
                                qi = sqrt( tanh(0.5*(loga - logzm))/(2*(loga - logzm)) )*(mim-I);
                            end
                            %St_dod = St_dod + qi*qi';
                            %------------------------------------------------------------------
                            Nq = qi;
                            tmp_tab1 = V'*Nq;
                            tmp_dd = V*tmp_tab1; 
                            Nq = Nq - tmp_dd;
                            S = S + tmp_tab1*tmp_tab1';
                            [Vq,Sq] = eig(S);
                            S = Sq;
                            V = V*Vq;               
                            norma_kw = norm(Nq)^2;
                            S_tmp = [diag(S); norma_kw];
                            [value,index] = min(S_tmp);
                            if(index <= k)
                                cc = S(index,index)*sum(abs(V(:,index)));
                                D_vec = D_vec + cc*(abs(V(:,index)));
                                S(index,index) = norma_kw;        
                                Nq = Nq/sqrt(norma_kw);
                                V(:,index) = Nq;
                            else
                                cc = sum(abs(Nq));
                                D_vec = D_vec + cc*abs(Nq);
                            end
                            %------------------------------------------------------------------
%                             comp1 = (a/(zm + a)).*I;
%                             comp2 = (zm/(zm + a)).*mim;
%                             mim = comp1 + comp2;
                            if(logzm>loga)
                                rat1 = 1.0/(1.0+exp(loga-logzm));
                                rat2 = 1.0-rat1;
                            else
                                rat2 = 1.0/(exp(logzm-loga)+1.0);
                                rat1 = 1.0-rat2;
                            end   
                            mim = rat1*mim + rat2*I;
                            %zm = zm + a;
                            if(loga<logzm) 
                                logzm = logzm + log(1.0+exp(loga-logzm));
                            else
                                logzm = loga + log(1.0+exp(logzm-loga));
                            end
                        end
                    end
                end     
            end
            mim = mim + LAMBDA*THETA;
            %zi = log(zm) + 0.5*LAMBDA*dot(THETA,THETA);
            zi=logzm+0.5*LAMBDA*dot(THETA,THETA);
            uti = uti+mim;
            zti = zti+zi;
            %S_dod = S_dod + St_dod;
        end        
        mim = uti-GG';
        zi = zti-dot(THETA,GG');
        %S_dod = (1/xi)*S_dod;
        %THETA = THETA-pinv(S_dod)*mim;
        tmp_dd = mim./D_vec;
        Vt = V';
        for ii=1:k
            Vt(ii,:) = Vt(ii,:)./(D_vec');
        end
        AA = Vt*V;
        AA = S*AA;
        AA = AA + eye(k,k);
        AA = pinv(AA);
        AA = AA*S;
        tmp_tab1 = Vt*tmp_dd;
        tmp_tab2 = AA*tmp_tab1;
        Vt = V';
        Nq = V*tmp_tab2;
        Nq = mim./D_vec - Nq./D_vec;
        THETA = THETA - xi*Nq;
        %Alternative update on THETA, however more computationally
        %expensive
        %THETA = THETA-xi*pinv(V*S*V' + diag(D_vec))*mim;
        RESULT_BOUND_Matlab_v1 = [RESULT_BOUND_Matlab_v1;zi,cputime-cpustart];
        save('RESULT_BOUND_Matlab_v1.mat','RESULT_BOUND_Matlab_v1');
        iter = iter + 1;
        disp(zi);
        disp(iter);
        %-----------------------Stopping criteria------------------------------
        if(iter > maxiter)
            disp('Optimization completed because the maximum number of iterations was reached');
            flag_iter = 0;
        end
        if(iter > 3)
            if((norm(RESULT_BOUND_Matlab_v1(iter)- RESULT_BOUND_Matlab_v1(iter-1)) < TolFun_val) && (norm(RESULT_BOUND_Matlab_v1(iter-1)- RESULT_BOUND_Matlab_v1(iter-2)) < TolFun_val) && (norm(RESULT_BOUND_Matlab_v1(iter-2)- RESULT_BOUND_Matlab_v1(iter-3)) < TolFun_val))
                disp('Optimization completed because the change of function value is less than the size of the function tolerance');
                flag_TolFun = 0;
            end
        end
    %     if(norm(THETA - theta_tmp) < TolX_val)
    %         disp('Optimization completed because the change of parameter vector is less than the size of the parameter tolerance');
    %         flag_TolX = 0;
    %     end
        %----------------------------------------------------------------------
    end
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_BOUND_Matlab_v1,1)
        if(RESULT_BOUND_Matlab_v1(i,1) > RESULT_BOUND_Matlab_v1(i-1,1))
            RESULT_BOUND_Matlab_v1(i,1) = RESULT_BOUND_Matlab_v1(i-1,1);
        end
    end
    %--------------------------------------------------------------------
end
%--------------------------------------------------------------------------

% Bound with sigma inversion xi = 1.0 Mex implementation (equivalent to Matlab implementation however faster)
%--------------------------------------------------------------------------
if(flag_BOUND_Mex_v2 == 1)
    k = 1;
    xi = 1.0;
    disp('Bound with sigma inversion xi = 1.0 Mex v.2');
    RESULT_BOUND_Mex_v2 = [];
    THETA = theta1;
    iter = 0;
    flag_iter = 1;
    flag_TolX = 1;
    flag_TolFun = 1;
    %------------------------------------------------------------------
    V_in = zeros(d,k);
    V_tmp = rand(k,k);
    [Q,R] = qr(V_tmp);
    V_in(1:k,1:k) = Q;
    D_vec_in = ones(d,1).*(t*LAMBDA);
    S_in = zeros(k,k);
   
    V = zeros(d,k);
    V_tmp = rand(k,k);
    [Q,R] = qr(V_tmp);
    V(1:k,1:k) = Q; 
    D_vec = ones(d,1).*(t*LAMBDA);
    S = zeros(k,k);
    %------------------------------------------------------------------
    cpustart = cputime;
    while(flag_iter && flag_TolX && flag_TolFun)       
        %------------------------------------------------------------------
        [zi,THETA,mim] = fast_k_bound(t,d,d_small,uint32(m),h,k,XXM,YYM,THETA,V,S,D_vec,V_in,S_in,D_vec_in,LAMBDA,xi);
        RESULT_BOUND_Mex_v2 = [RESULT_BOUND_Mex_v2;zi,cputime-cpustart,THETA'];
        save('RESULT_BOUND_Mex_v2.mat','RESULT_BOUND_Mex_v2');
        iter = iter + 1;
        disp(zi);
        disp(iter);
        %-----------------------Stopping criteria------------------------------
        if(iter > maxiter)
            disp('Optimization completed because the maximum number of iterations was reached');
            flag_iter = 0;
        end
        if(iter > 3)
            if((norm(RESULT_BOUND_Mex_v2(iter)- RESULT_BOUND_Mex_v2(iter-1)) < TolFun_val) && (norm(RESULT_BOUND_Mex_v2(iter-1)- RESULT_BOUND_Mex_v2(iter-2)) < TolFun_val) && (norm(RESULT_BOUND_Mex_v2(iter-2)- RESULT_BOUND_Mex_v2(iter-3)) < TolFun_val))
                disp('Optimization completed because the change of function value is less than the size of the function tolerance');
                flag_TolFun = 0;
            end
        end
    %     if(norm(THETA - theta_tmp) < TolX_val)
    %         disp('Optimization completed because the change of parameter vector is less than the size of the parameter tolerance');
    %         flag_TolX = 0;
    %     end
        %----------------------------------------------------------------------
    end
    %curves smoothing
    %--------------------------------------------------------------------
    for i = 2:1:size(RESULT_BOUND_Mex_v2,1)
        if(RESULT_BOUND_Mex_v2(i,1) > RESULT_BOUND_Mex_v2(i-1,1))
            RESULT_BOUND_Mex_v2(i,1) = RESULT_BOUND_Mex_v2(i-1,1);
        end
    end
    %--------------------------------------------------------------------
end
%--------------------------------------------------------------------------
