% Code by Anna Choromanska and Tony Jebara
% Reference: T. Jebara and A. Choromanska, "Majorization for CRFs and Latent Likelihoods", Neural Information Processing Systems (NIPS), December 2012. 

clear all
close all
clc

% GLOBAL VARIABLES
global flag_theta_gen;
global flag_LBFGS;
global flag_SD;
global flag_CG;
global flag_BOUND_Matlab_v1;
global flag_BOUND_Mex_v2;

% FLAGS
flag_LBFGS = 1;
flag_SD = 1;
flag_CG = 1;
flag_BOUND_Matlab_v1 = 0;
flag_BOUND_Mex_v2 = 1;
flag_theta_gen = 0;

% VARIABLES
start_fi = 1;
stop_fi = 1;
start_ll = 1;
stop_ll = 1;
start_iter = 1;
stop_iter = 1;
nbofiter = stop_iter - start_iter + 1;
nbofmethods = flag_LBFGS + flag_SD + flag_CG + flag_BOUND_Matlab_v1 + flag_BOUND_Mex_v2;
reg = [3];
maxiter = 100000;
TolX_val = 1e-6;
TolFun_val = 1e-6;
p = 0.05;
 
load('data.mat');

% TO SEE SAVED RESULTS, COMMENT THE JTA PART BELOW AND RUN script_LR
% JTA
for file = start_fi:1:stop_fi
    file_now = file;
    disp(file_now); 
    for ll=start_ll:1:stop_ll
        disp(ll);
        for iteration = start_iter:1:stop_iter
            disp(iteration);
            lamb=10^reg(ll);
            [RESULT_LBFGS,RESULT_SD,RESULT_CG,RESULT_BOUND_Matlab_v1,RESULT_BOUND_Mex_v2] = realbound(xxx,yyy,lamb,maxiter,TolX_val,TolFun_val,file_now,ll,iteration);
            s=sprintf('Trees%d.%d.%d.mat',file_now,ll,iteration); save(s,'RESULT_LBFGS','RESULT_SD','RESULT_CG','RESULT_BOUND_Matlab_v1','RESULT_BOUND_Mex_v2');
        end
    end
end

%RESULTS ANALYSIS
for file=start_fi:1:stop_fi
    file_now = file;
    for ll=start_ll:1:stop_ll
        time(:,file_now,ll) = zeros(nbofmethods,1); 
        value(:,file_now,ll) = zeros(nbofmethods,1);
        iterations(:,file_now,ll) = zeros(nbofmethods,1);
        value_t(:,file_now,ll) = zeros(nbofmethods,1);
        percent(:,file_now,ll) = zeros(nbofmethods,1);
        percent_t(:,file_now,ll) = zeros(nbofmethods,1);
        value_conv(:,file_now,ll) = zeros(nbofmethods,1);
        time_p(:,file_now,ll) = zeros(nbofmethods,1);
        iterations_p(:,file_now,ll) = zeros(nbofmethods,1);
        value_p(:,file_now,ll) = zeros(nbofmethods,1);
        value_t_p(:,file_now,ll) = zeros(nbofmethods,1);
        for iteration = 1:1:nbofiter
            FINAL_TAB = [];
            [FINAL_TAB] = analysis(maxiter,p,file_now,ll,iteration);
            for  i = 1:1:nbofmethods
                value(i,file_now,ll) = value(i,file_now,ll) + FINAL_TAB(i,1);
                time(i,file_now,ll) = time(i,file_now,ll) + FINAL_TAB(i,2);
                iterations(i,file_now,ll) = iterations(i,file_now,ll) + FINAL_TAB(i,3);
                value_t(i,file_now,ll) = value_t(i,file_now,ll) + FINAL_TAB(i,4);
                percent(i,file_now,ll) = percent(i,file_now,ll) + FINAL_TAB(i,5);
                percent_t(i,file_now,ll) = percent_t(i,file_now,ll) + FINAL_TAB(i,6);
                value_conv(i,file_now,ll) = value_conv(i,file_now,ll) + FINAL_TAB(i,7);
                time_p(i,file_now,ll) = time_p(i,file_now,ll) + FINAL_TAB(i,8);
                iterations_p(i,file_now,ll) = iterations_p(i,file_now,ll) + FINAL_TAB(i,9);
                value_p(i,file_now,ll) = value_p(i,file_now,ll) + FINAL_TAB(i,10);
                value_t_p(i,file_now,ll) = value_t_p(i,file_now,ll) + FINAL_TAB(i,11);
            end
        end
        time(:,file_now,ll) = time(:,file_now,ll)/nbofiter;
        value(:,file_now,ll) = value(:,file_now,ll)/nbofiter;
        iterations(:,file_now,ll) = round(iterations(:,file_now,ll)/nbofiter);
        value_t(:,file_now,ll) = value_t(:,file_now,ll)/nbofiter;
        percent(:,file_now,ll) = percent(:,file_now,ll)/nbofiter;
        percent_t(:,file_now,ll) = percent_t(:,file_now,ll)/nbofiter;
        value_conv(:,file_now,ll) = value_conv(:,file_now,ll)/nbofiter;
        time_p(:,file_now,ll) = time_p(:,file_now,ll)/nbofiter;
        iterations_p(:,file_now,ll) = iterations_p(:,file_now,ll)/nbofiter;
        value_p(:,file_now,ll) = value_p(:,file_now,ll)/nbofiter;
        value_t_p(:,file_now,ll) = value_t_p(:,file_now,ll)/nbofiter;
    end
end

if(flag_LBFGS == 1) fprintf('LBFGS');fprintf(1,' \n'); end
if(flag_SD == 1) fprintf('SD');fprintf(1,' \n'); end
if(flag_CG == 1) fprintf('CG');fprintf(1,' \n'); end
if(flag_BOUND_Mex_v2 == 1) fprintf('BOUND Mex v.2');fprintf(1,' \n'); end
fprintf(1,' \n');
fprintf(1,' \n');

% RESULTS DISPLAY
if(flag_LBFGS == 1) fprintf('LBFGS');fprintf(1,' \n'); end
if(flag_SD == 1) fprintf('SD');fprintf(1,' \n'); end
if(flag_CG == 1) fprintf('CG');fprintf(1,' \n'); end
if(flag_BOUND_Matlab_v1 == 1) fprintf('BOUND Matlab v.1');fprintf(1,' \n'); end
if(flag_BOUND_Mex_v2 == 1) fprintf('BOUND Mex v.2');fprintf(1,' \n'); end
fprintf(1,' \n');
fprintf(1,' \n');

fprintf(1,' \n');
disp('time');
for  i = 1:1:nbofmethods
    for file=start_fi:1:stop_fi
        file_now = file;
        for ll=start_ll:1:stop_ll
            fprintf(1,'%2.2f & ', time(i,file_now,ll));
        end
    end
    fprintf(1,' \n');
end
fprintf(1,' \n');

disp('iterations');
for  i = 1:1:nbofmethods
    for file=start_fi:1:stop_fi
        file_now = file;
        for ll=start_ll:1:stop_ll
            fprintf(1,'%2.0f & ', iterations(i,file_now,ll));
        end
    end
    fprintf(1,' \n');
end
fprintf(1,' \n');

fprintf(1,' \n');
disp('value');
for  i = 1:1:nbofmethods
    for file=start_fi:1:stop_fi
        file_now = file;
        for ll=start_ll:1:stop_ll
            fprintf(1,'%2.1f & ', (-1)*value(i,file_now,ll));
        end
    end
    fprintf(1,' \n');
end
fprintf(1,' \n');
